# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: v1/aggregate.proto
# Protobuf Python Version: 6.30.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    30,
    0,
    '',
    'v1/aggregate.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from weaviate.proto.v1.v6300.v1 import base_pb2 as v1_dot_base__pb2
from weaviate.proto.v1.v6300.v1 import base_search_pb2 as v1_dot_base__search__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12v1/aggregate.proto\x12\x0bweaviate.v1\x1a\rv1/base.proto\x1a\x14v1/base_search.proto\"\xe8\x0f\n\x10\x41ggregateRequest\x12\x12\n\ncollection\x18\x01 \x01(\t\x12\x0e\n\x06tenant\x18\n \x01(\t\x12\x15\n\robjects_count\x18\x14 \x01(\x08\x12?\n\x0c\x61ggregations\x18\x15 \x03(\x0b\x32).weaviate.v1.AggregateRequest.Aggregation\x12\x19\n\x0cobject_limit\x18\x1e \x01(\rH\x01\x88\x01\x01\x12<\n\x08group_by\x18\x1f \x01(\x0b\x32%.weaviate.v1.AggregateRequest.GroupByH\x02\x88\x01\x01\x12\x12\n\x05limit\x18  \x01(\rH\x03\x88\x01\x01\x12*\n\x07\x66ilters\x18( \x01(\x0b\x32\x14.weaviate.v1.FiltersH\x04\x88\x01\x01\x12%\n\x06hybrid\x18) \x01(\x0b\x32\x13.weaviate.v1.HybridH\x00\x12.\n\x0bnear_vector\x18* \x01(\x0b\x32\x17.weaviate.v1.NearVectorH\x00\x12.\n\x0bnear_object\x18+ \x01(\x0b\x32\x17.weaviate.v1.NearObjectH\x00\x12\x30\n\tnear_text\x18, \x01(\x0b\x32\x1b.weaviate.v1.NearTextSearchH\x00\x12\x32\n\nnear_image\x18- \x01(\x0b\x32\x1c.weaviate.v1.NearImageSearchH\x00\x12\x32\n\nnear_audio\x18. \x01(\x0b\x32\x1c.weaviate.v1.NearAudioSearchH\x00\x12\x32\n\nnear_video\x18/ \x01(\x0b\x32\x1c.weaviate.v1.NearVideoSearchH\x00\x12\x32\n\nnear_depth\x18\x30 \x01(\x0b\x32\x1c.weaviate.v1.NearDepthSearchH\x00\x12\x36\n\x0cnear_thermal\x18\x31 \x01(\x0b\x32\x1e.weaviate.v1.NearThermalSearchH\x00\x12.\n\x08near_imu\x18\x32 \x01(\x0b\x32\x1a.weaviate.v1.NearIMUSearchH\x00\x1a\xde\x08\n\x0b\x41ggregation\x12\x10\n\x08property\x18\x01 \x01(\t\x12@\n\x03int\x18\x02 \x01(\x0b\x32\x31.weaviate.v1.AggregateRequest.Aggregation.IntegerH\x00\x12\x42\n\x06number\x18\x03 \x01(\x0b\x32\x30.weaviate.v1.AggregateRequest.Aggregation.NumberH\x00\x12>\n\x04text\x18\x04 \x01(\x0b\x32..weaviate.v1.AggregateRequest.Aggregation.TextH\x00\x12\x44\n\x07\x62oolean\x18\x05 \x01(\x0b\x32\x31.weaviate.v1.AggregateRequest.Aggregation.BooleanH\x00\x12>\n\x04\x64\x61te\x18\x06 \x01(\x0b\x32..weaviate.v1.AggregateRequest.Aggregation.DateH\x00\x12H\n\treference\x18\x07 \x01(\x0b\x32\x33.weaviate.v1.AggregateRequest.Aggregation.ReferenceH\x00\x1a\x81\x01\n\x07Integer\x12\r\n\x05\x63ount\x18\x01 \x01(\x08\x12\x0c\n\x04type\x18\x02 \x01(\x08\x12\x0b\n\x03sum\x18\x03 \x01(\x08\x12\x0c\n\x04mean\x18\x04 \x01(\x08\x12\x0c\n\x04mode\x18\x05 \x01(\x08\x12\x0e\n\x06median\x18\x06 \x01(\x08\x12\x0f\n\x07maximum\x18\x07 \x01(\x08\x12\x0f\n\x07minimum\x18\x08 \x01(\x08\x1a\x80\x01\n\x06Number\x12\r\n\x05\x63ount\x18\x01 \x01(\x08\x12\x0c\n\x04type\x18\x02 \x01(\x08\x12\x0b\n\x03sum\x18\x03 \x01(\x08\x12\x0c\n\x04mean\x18\x04 \x01(\x08\x12\x0c\n\x04mode\x18\x05 \x01(\x08\x12\x0e\n\x06median\x18\x06 \x01(\x08\x12\x0f\n\x07maximum\x18\x07 \x01(\x08\x12\x0f\n\x07minimum\x18\x08 \x01(\x08\x1aw\n\x04Text\x12\r\n\x05\x63ount\x18\x01 \x01(\x08\x12\x0c\n\x04type\x18\x02 \x01(\x08\x12\x16\n\x0etop_occurences\x18\x03 \x01(\x08\x12!\n\x14top_occurences_limit\x18\x04 \x01(\rH\x00\x88\x01\x01\x42\x17\n\x15_top_occurences_limit\x1a\x82\x01\n\x07\x42oolean\x12\r\n\x05\x63ount\x18\x01 \x01(\x08\x12\x0c\n\x04type\x18\x02 \x01(\x08\x12\x12\n\ntotal_true\x18\x03 \x01(\x08\x12\x13\n\x0btotal_false\x18\x04 \x01(\x08\x12\x17\n\x0fpercentage_true\x18\x05 \x01(\x08\x12\x18\n\x10percentage_false\x18\x06 \x01(\x08\x1a\x63\n\x04\x44\x61te\x12\r\n\x05\x63ount\x18\x01 \x01(\x08\x12\x0c\n\x04type\x18\x02 \x01(\x08\x12\x0e\n\x06median\x18\x03 \x01(\x08\x12\x0c\n\x04mode\x18\x04 \x01(\x08\x12\x0f\n\x07maximum\x18\x05 \x01(\x08\x12\x0f\n\x07minimum\x18\x06 \x01(\x08\x1a.\n\tReference\x12\x0c\n\x04type\x18\x01 \x01(\x08\x12\x13\n\x0bpointing_to\x18\x02 \x01(\x08\x42\r\n\x0b\x61ggregation\x1a/\n\x07GroupBy\x12\x12\n\ncollection\x18\x01 \x01(\t\x12\x10\n\x08property\x18\x02 \x01(\tB\x08\n\x06searchB\x0f\n\r_object_limitB\x0b\n\t_group_byB\x08\n\x06_limitB\n\n\x08_filters\"\xdf\x16\n\x0e\x41ggregateReply\x12\x0c\n\x04took\x18\x01 \x01(\x02\x12;\n\rsingle_result\x18\x02 \x01(\x0b\x32\".weaviate.v1.AggregateReply.SingleH\x00\x12>\n\x0fgrouped_results\x18\x03 \x01(\x0b\x32#.weaviate.v1.AggregateReply.GroupedH\x00\x1a\xc0\x0f\n\x0c\x41ggregations\x12J\n\x0c\x61ggregations\x18\x01 \x03(\x0b\x32\x34.weaviate.v1.AggregateReply.Aggregations.Aggregation\x1a\xe3\x0e\n\x0b\x41ggregation\x12\x10\n\x08property\x18\x01 \x01(\t\x12K\n\x03int\x18\x02 \x01(\x0b\x32<.weaviate.v1.AggregateReply.Aggregations.Aggregation.IntegerH\x00\x12M\n\x06number\x18\x03 \x01(\x0b\x32;.weaviate.v1.AggregateReply.Aggregations.Aggregation.NumberH\x00\x12I\n\x04text\x18\x04 \x01(\x0b\x32\x39.weaviate.v1.AggregateReply.Aggregations.Aggregation.TextH\x00\x12O\n\x07\x62oolean\x18\x05 \x01(\x0b\x32<.weaviate.v1.AggregateReply.Aggregations.Aggregation.BooleanH\x00\x12I\n\x04\x64\x61te\x18\x06 \x01(\x0b\x32\x39.weaviate.v1.AggregateReply.Aggregations.Aggregation.DateH\x00\x12S\n\treference\x18\x07 \x01(\x0b\x32>.weaviate.v1.AggregateReply.Aggregations.Aggregation.ReferenceH\x00\x1a\xf9\x01\n\x07Integer\x12\x12\n\x05\x63ount\x18\x01 \x01(\x03H\x00\x88\x01\x01\x12\x11\n\x04type\x18\x02 \x01(\tH\x01\x88\x01\x01\x12\x11\n\x04mean\x18\x03 \x01(\x01H\x02\x88\x01\x01\x12\x13\n\x06median\x18\x04 \x01(\x01H\x03\x88\x01\x01\x12\x11\n\x04mode\x18\x05 \x01(\x03H\x04\x88\x01\x01\x12\x14\n\x07maximum\x18\x06 \x01(\x03H\x05\x88\x01\x01\x12\x14\n\x07minimum\x18\x07 \x01(\x03H\x06\x88\x01\x01\x12\x10\n\x03sum\x18\x08 \x01(\x03H\x07\x88\x01\x01\x42\x08\n\x06_countB\x07\n\x05_typeB\x07\n\x05_meanB\t\n\x07_medianB\x07\n\x05_modeB\n\n\x08_maximumB\n\n\x08_minimumB\x06\n\x04_sum\x1a\xf8\x01\n\x06Number\x12\x12\n\x05\x63ount\x18\x01 \x01(\x03H\x00\x88\x01\x01\x12\x11\n\x04type\x18\x02 \x01(\tH\x01\x88\x01\x01\x12\x11\n\x04mean\x18\x03 \x01(\x01H\x02\x88\x01\x01\x12\x13\n\x06median\x18\x04 \x01(\x01H\x03\x88\x01\x01\x12\x11\n\x04mode\x18\x05 \x01(\x01H\x04\x88\x01\x01\x12\x14\n\x07maximum\x18\x06 \x01(\x01H\x05\x88\x01\x01\x12\x14\n\x07minimum\x18\x07 \x01(\x01H\x06\x88\x01\x01\x12\x10\n\x03sum\x18\x08 \x01(\x01H\x07\x88\x01\x01\x42\x08\n\x06_countB\x07\n\x05_typeB\x07\n\x05_meanB\t\n\x07_medianB\x07\n\x05_modeB\n\n\x08_maximumB\n\n\x08_minimumB\x06\n\x04_sum\x1a\xe4\x02\n\x04Text\x12\x12\n\x05\x63ount\x18\x01 \x01(\x03H\x00\x88\x01\x01\x12\x11\n\x04type\x18\x02 \x01(\tH\x01\x88\x01\x01\x12\x65\n\x0etop_occurences\x18\x03 \x01(\x0b\x32H.weaviate.v1.AggregateReply.Aggregations.Aggregation.Text.TopOccurrencesH\x02\x88\x01\x01\x1a\xa7\x01\n\x0eTopOccurrences\x12\x65\n\x05items\x18\x01 \x03(\x0b\x32V.weaviate.v1.AggregateReply.Aggregations.Aggregation.Text.TopOccurrences.TopOccurrence\x1a.\n\rTopOccurrence\x12\r\n\x05value\x18\x01 \x01(\t\x12\x0e\n\x06occurs\x18\x02 \x01(\x03\x42\x08\n\x06_countB\x07\n\x05_typeB\x11\n\x0f_top_occurences\x1a\xfb\x01\n\x07\x42oolean\x12\x12\n\x05\x63ount\x18\x01 \x01(\x03H\x00\x88\x01\x01\x12\x11\n\x04type\x18\x02 \x01(\tH\x01\x88\x01\x01\x12\x17\n\ntotal_true\x18\x03 \x01(\x03H\x02\x88\x01\x01\x12\x18\n\x0btotal_false\x18\x04 \x01(\x03H\x03\x88\x01\x01\x12\x1c\n\x0fpercentage_true\x18\x05 \x01(\x01H\x04\x88\x01\x01\x12\x1d\n\x10percentage_false\x18\x06 \x01(\x01H\x05\x88\x01\x01\x42\x08\n\x06_countB\x07\n\x05_typeB\r\n\x0b_total_trueB\x0e\n\x0c_total_falseB\x12\n\x10_percentage_trueB\x13\n\x11_percentage_false\x1a\xc0\x01\n\x04\x44\x61te\x12\x12\n\x05\x63ount\x18\x01 \x01(\x03H\x00\x88\x01\x01\x12\x11\n\x04type\x18\x02 \x01(\tH\x01\x88\x01\x01\x12\x13\n\x06median\x18\x03 \x01(\tH\x02\x88\x01\x01\x12\x11\n\x04mode\x18\x04 \x01(\tH\x03\x88\x01\x01\x12\x14\n\x07maximum\x18\x05 \x01(\tH\x04\x88\x01\x01\x12\x14\n\x07minimum\x18\x06 \x01(\tH\x05\x88\x01\x01\x42\x08\n\x06_countB\x07\n\x05_typeB\t\n\x07_medianB\x07\n\x05_modeB\n\n\x08_maximumB\n\n\x08_minimum\x1a<\n\tReference\x12\x11\n\x04type\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x13\n\x0bpointing_to\x18\x02 \x03(\tB\x07\n\x05_typeB\r\n\x0b\x61ggregation\x1a\x8c\x01\n\x06Single\x12\x1a\n\robjects_count\x18\x01 \x01(\x03H\x00\x88\x01\x01\x12\x43\n\x0c\x61ggregations\x18\x02 \x01(\x0b\x32(.weaviate.v1.AggregateReply.AggregationsH\x01\x88\x01\x01\x42\x10\n\x0e_objects_countB\x0f\n\r_aggregations\x1a\xa7\x04\n\x05Group\x12\x1a\n\robjects_count\x18\x01 \x01(\x03H\x00\x88\x01\x01\x12\x43\n\x0c\x61ggregations\x18\x02 \x01(\x0b\x32(.weaviate.v1.AggregateReply.AggregationsH\x01\x88\x01\x01\x12\x44\n\ngrouped_by\x18\x03 \x01(\x0b\x32+.weaviate.v1.AggregateReply.Group.GroupedByH\x02\x88\x01\x01\x1a\xc4\x02\n\tGroupedBy\x12\x0c\n\x04path\x18\x01 \x03(\t\x12\x0e\n\x04text\x18\x02 \x01(\tH\x00\x12\r\n\x03int\x18\x03 \x01(\x03H\x00\x12\x11\n\x07\x62oolean\x18\x04 \x01(\x08H\x00\x12\x10\n\x06number\x18\x05 \x01(\x01H\x00\x12\'\n\x05texts\x18\x06 \x01(\x0b\x32\x16.weaviate.v1.TextArrayH\x00\x12%\n\x04ints\x18\x07 \x01(\x0b\x32\x15.weaviate.v1.IntArrayH\x00\x12-\n\x08\x62ooleans\x18\x08 \x01(\x0b\x32\x19.weaviate.v1.BooleanArrayH\x00\x12+\n\x07numbers\x18\t \x01(\x0b\x32\x18.weaviate.v1.NumberArrayH\x00\x12\x30\n\x03geo\x18\n \x01(\x0b\x32!.weaviate.v1.GeoCoordinatesFilterH\x00\x42\x07\n\x05valueB\x10\n\x0e_objects_countB\x0f\n\r_aggregationsB\r\n\x0b_grouped_by\x1a<\n\x07Grouped\x12\x31\n\x06groups\x18\x01 \x03(\x0b\x32!.weaviate.v1.AggregateReply.GroupB\x08\n\x06resultBs\n#io.weaviate.client.grpc.protocol.v1B\x16WeaviateProtoAggregateZ4github.com/weaviate/weaviate/grpc/generated;protocolb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'v1.aggregate_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n#io.weaviate.client.grpc.protocol.v1B\026WeaviateProtoAggregateZ4github.com/weaviate/weaviate/grpc/generated;protocol'
  _globals['_AGGREGATEREQUEST']._serialized_start=73
  _globals['_AGGREGATEREQUEST']._serialized_end=2097
  _globals['_AGGREGATEREQUEST_AGGREGATION']._serialized_start=868
  _globals['_AGGREGATEREQUEST_AGGREGATION']._serialized_end=1986
  _globals['_AGGREGATEREQUEST_AGGREGATION_INTEGER']._serialized_start=1308
  _globals['_AGGREGATEREQUEST_AGGREGATION_INTEGER']._serialized_end=1437
  _globals['_AGGREGATEREQUEST_AGGREGATION_NUMBER']._serialized_start=1440
  _globals['_AGGREGATEREQUEST_AGGREGATION_NUMBER']._serialized_end=1568
  _globals['_AGGREGATEREQUEST_AGGREGATION_TEXT']._serialized_start=1570
  _globals['_AGGREGATEREQUEST_AGGREGATION_TEXT']._serialized_end=1689
  _globals['_AGGREGATEREQUEST_AGGREGATION_BOOLEAN']._serialized_start=1692
  _globals['_AGGREGATEREQUEST_AGGREGATION_BOOLEAN']._serialized_end=1822
  _globals['_AGGREGATEREQUEST_AGGREGATION_DATE']._serialized_start=1824
  _globals['_AGGREGATEREQUEST_AGGREGATION_DATE']._serialized_end=1923
  _globals['_AGGREGATEREQUEST_AGGREGATION_REFERENCE']._serialized_start=1925
  _globals['_AGGREGATEREQUEST_AGGREGATION_REFERENCE']._serialized_end=1971
  _globals['_AGGREGATEREQUEST_GROUPBY']._serialized_start=1988
  _globals['_AGGREGATEREQUEST_GROUPBY']._serialized_end=2035
  _globals['_AGGREGATEREPLY']._serialized_start=2100
  _globals['_AGGREGATEREPLY']._serialized_end=5011
  _globals['_AGGREGATEREPLY_AGGREGATIONS']._serialized_start=2258
  _globals['_AGGREGATEREPLY_AGGREGATIONS']._serialized_end=4242
  _globals['_AGGREGATEREPLY_AGGREGATIONS_AGGREGATION']._serialized_start=2351
  _globals['_AGGREGATEREPLY_AGGREGATIONS_AGGREGATION']._serialized_end=4242
  _globals['_AGGREGATEREPLY_AGGREGATIONS_AGGREGATION_INTEGER']._serialized_start=2857
  _globals['_AGGREGATEREPLY_AGGREGATIONS_AGGREGATION_INTEGER']._serialized_end=3106
  _globals['_AGGREGATEREPLY_AGGREGATIONS_AGGREGATION_NUMBER']._serialized_start=3109
  _globals['_AGGREGATEREPLY_AGGREGATIONS_AGGREGATION_NUMBER']._serialized_end=3357
  _globals['_AGGREGATEREPLY_AGGREGATIONS_AGGREGATION_TEXT']._serialized_start=3360
  _globals['_AGGREGATEREPLY_AGGREGATIONS_AGGREGATION_TEXT']._serialized_end=3716
  _globals['_AGGREGATEREPLY_AGGREGATIONS_AGGREGATION_TEXT_TOPOCCURRENCES']._serialized_start=3511
  _globals['_AGGREGATEREPLY_AGGREGATIONS_AGGREGATION_TEXT_TOPOCCURRENCES']._serialized_end=3678
  _globals['_AGGREGATEREPLY_AGGREGATIONS_AGGREGATION_TEXT_TOPOCCURRENCES_TOPOCCURRENCE']._serialized_start=3632
  _globals['_AGGREGATEREPLY_AGGREGATIONS_AGGREGATION_TEXT_TOPOCCURRENCES_TOPOCCURRENCE']._serialized_end=3678
  _globals['_AGGREGATEREPLY_AGGREGATIONS_AGGREGATION_BOOLEAN']._serialized_start=3719
  _globals['_AGGREGATEREPLY_AGGREGATIONS_AGGREGATION_BOOLEAN']._serialized_end=3970
  _globals['_AGGREGATEREPLY_AGGREGATIONS_AGGREGATION_DATE']._serialized_start=3973
  _globals['_AGGREGATEREPLY_AGGREGATIONS_AGGREGATION_DATE']._serialized_end=4165
  _globals['_AGGREGATEREPLY_AGGREGATIONS_AGGREGATION_REFERENCE']._serialized_start=4167
  _globals['_AGGREGATEREPLY_AGGREGATIONS_AGGREGATION_REFERENCE']._serialized_end=4227
  _globals['_AGGREGATEREPLY_SINGLE']._serialized_start=4245
  _globals['_AGGREGATEREPLY_SINGLE']._serialized_end=4385
  _globals['_AGGREGATEREPLY_GROUP']._serialized_start=4388
  _globals['_AGGREGATEREPLY_GROUP']._serialized_end=4939
  _globals['_AGGREGATEREPLY_GROUP_GROUPEDBY']._serialized_start=4565
  _globals['_AGGREGATEREPLY_GROUP_GROUPEDBY']._serialized_end=4889
  _globals['_AGGREGATEREPLY_GROUPED']._serialized_start=4941
  _globals['_AGGREGATEREPLY_GROUPED']._serialized_end=5001
# @@protoc_insertion_point(module_scope)



from typing import Generic

from weaviate.collections.classes.types import Properties, References
from weaviate.collections.queries.fetch_objects_by_ids.query.executor import (
    _FetchObjectsByIDsQueryExecutor,
)
from weaviate.connect import executor
from weaviate.connect.v4 import ConnectionSync


@executor.wrap("sync")
class _FetchObjectsByIDsQuery(
    Generic[Properties, References],
    _FetchObjectsByIDsQueryExecutor[ConnectionSync, Properties, References],
):
    pass



import ast
import importlib
import inspect
import os
import textwrap
from collections import defaultdict
from typing import Literal, cast


class ExecutorTransformer(ast.NodeTransformer):
    def __init__(self, colour: Literal["async", "sync"]):
        self.colour = colour
        self.executor_names = []

    def visit_ClassDef(self, node):
        self.executor_names.append(node.name)
        node.bases = self.__parse_generics(node)
        node.body = self.__parse_body(node)
        node.name = node.name.replace(
            "Executor", "" if self.colour == "sync" else self.colour.capitalize()
        )
        self.generic_visit(node)
        return node

    def __is_overload(self, fn: ast.FunctionDef):
        return any(isinstance(d, ast.Name) and d.id == "overload" for d in fn.decorator_list)

    def __parse_body(self, node: ast.ClassDef):
        funcs_by_name = defaultdict(list)
        for stmt in node.body:
            if isinstance(stmt, (ast.FunctionDef, ast.AsyncFunctionDef)):
                funcs_by_name[stmt.name].append(stmt)

        new_body: list[ast.stmt] = []
        for stmt in node.body:
            if isinstance(stmt, ast.FunctionDef) and stmt.name.startswith("__"):
                continue  # Skip all dunder methods
            if isinstance(stmt, ast.FunctionDef):
                overloads = funcs_by_name[stmt.name]
                if any(self.__is_overload(f) for f in overloads):
                    if not self.__is_overload(stmt):
                        continue  # skip the impl
            new_body.append(stmt)
        return new_body

    def __parse_generics(self, node: ast.ClassDef):
        new_bases: list[ast.expr] = []
        for base in node.bases:
            if not isinstance(base, ast.Subscript):
                continue
            if isinstance(base.value, ast.Name) and base.value.id == "Generic":
                # This is a generic class
                # We need to extract the type arguments
                if isinstance(base.slice, ast.Tuple):
                    # This is a tuple of types
                    # must remove `ConnectionType` if there
                    generics = [
                        arg.id
                        for arg in base.slice.elts
                        if isinstance(arg, ast.Name)
                        if arg.id != "ConnectionType"
                    ]
                    new_bases.append(
                        ast.Subscript(
                            value=base.value,
                            slice=ast.Tuple(
                                elts=[ast.Name(id=arg) for arg in generics], ctx=ast.Load()
                            ),
                            ctx=ast.Load(),
                        )
                    )
                elif isinstance(base.slice, ast.Name):
                    # This is a single type
                    if base.slice.id == "ConnectionType":
                        # We don't want to include ConnectionType
                        continue
                    new_bases.append(base)
        connection_type = ast.Name(id=self.__which_connection_type(), ctx=ast.Load())
        if len(new_bases) == 0:
            # no generics, we need to add the ConnectionType
            slice = connection_type
        else:
            elts: list[ast.expr] = []
            for base in new_bases:
                assert isinstance(base, ast.Subscript)
                slice = base.slice
                assert isinstance(slice, ast.Tuple)
                elts.extend(slice.elts)
            slice = ast.Tuple(elts=[connection_type, *elts], ctx=ast.Load())
        new_bases.append(
            ast.Subscript(
                value=ast.Name(id=node.name, ctx=ast.Load()),
                slice=slice,
                ctx=ast.Load(),
            )
        )
        return new_bases

    def __which_connection_type(self):
        return "ConnectionAsync" if self.colour == "async" else "ConnectionSync"

    def __extract_inner_return_type(self, node: ast.expr | None) -> ast.expr | None:
        # Looking for executor.Result[T]
        if (
            isinstance(node, ast.Subscript)
            and isinstance(node.value, ast.Attribute)
            and isinstance(node.value.value, ast.Name)
            and node.value.value.id == "executor"
            and node.value.attr == "Result"
        ):
            # This is executor.Result[...]
            return node.slice  # Return T
        return node  # fallback, return original if not matching

    def visit_FunctionDef(self, node):
        func_def = ast.AsyncFunctionDef if self.colour == "async" else ast.FunctionDef
        new_node = func_def(
            name=node.name,
            args=node.args,
            body=[ast.Expr(value=ast.Constant(value=Ellipsis))],
            decorator_list=node.decorator_list,
            returns=self.__extract_inner_return_type(node.returns),
            type_comment=node.type_comment,
        )
        return ast.copy_location(new_node, node)


for subdir, dirs, files in os.walk("./weaviate"):
    for file in files:
        if file != "executor.py":
            continue
        if "connect" in subdir:
            # ignore weaviate/connect/executor.py file
            continue
        if "collections/collections" in subdir:
            # ignore weaviate/collections/collections directory
            continue

        mod = os.path.join(subdir, file)
        mod = mod[2:]  # remove the leading dot and slash
        mod = mod[:-3]  # remove the .py
        mod = mod.replace("/", ".")  # convert into pythonic import

        module = importlib.import_module(mod)
        source = textwrap.dedent(inspect.getsource(module))

        colours: list[Literal["sync", "async"]] = ["sync", "async"]
        for colour in colours:
            tree = ast.parse(source, mode="exec", type_comments=True)

            transformer = ExecutorTransformer(colour)
            stubbed = transformer.visit(tree)

            imports = [
                node for node in stubbed.body if isinstance(node, (ast.Import, ast.ImportFrom))
            ] + [
                ast.ImportFrom(
                    module="weaviate.connect.v4",
                    names=[ast.alias(name=f"Connection{colour.capitalize()}", asname=None)],
                    level=0,
                ),
                ast.ImportFrom(
                    module=".executor",
                    names=[
                        ast.alias(name=name, asname=None) for name in transformer.executor_names
                    ],
                    level=0,
                ),
            ]
            stubbed.body = imports + [
                node for node in stubbed.body if isinstance(node, ast.ClassDef)
            ]
            ast.fix_missing_locations(stubbed)

            dir = cast(str, module.__package__).replace(".", "/")
            file = f"{dir}/{colour}.pyi" if colour == "sync" else f"{dir}/{colour}_.pyi"
            with open(file, "w") as f:
                print(f"Writing {file}")
                f.write(ast.unparse(stubbed))



from weaviate.collections.tenants.executor import _TenantsExecutor
from weaviate.connect import executor
from weaviate.connect.v4 import ConnectionAsync


@executor.wrap("async")
class _TenantsAsync(_TenantsExecutor[ConnectionAsync]):
    pass



from weaviate.connect import executor
from weaviate.connect.v4 import ConnectionSync
from weaviate.users.base import (
    _UsersDBExecutor,
    _UsersExecutor,
    _UsersOIDCExecutor,
)


@executor.wrap("sync")
class _UsersDB(_UsersDBExecutor[ConnectionSync]):
    pass


@executor.wrap("sync")
class _UsersOIDC(_UsersOIDCExecutor[ConnectionSync]):
    pass


@executor.wrap("sync")
class _Users(_UsersExecutor[ConnectionSync]):
    def __init__(self, connection: ConnectionSync):
        super().__init__(connection)
        self.db = _UsersDB(connection)
        self.oidc = _UsersOIDC(connection)



from datetime import datetime
from typing import Any, Dict, Optional

from pydantic import BaseModel, Field

from weaviate.types import uuid_package


class DebugRESTObject(BaseModel):
    collection: str = Field(..., alias="class")
    creation_time: datetime = Field(..., alias="creationTimeUnix")
    last_update_time: datetime = Field(..., alias="lastUpdateTimeUnix")
    properties: Dict[str, Any] = Field(...)
    tenant: Optional[str] = Field(None)
    uuid: uuid_package.UUID = Field(..., alias="id")
    vector: Optional[list[float]] = Field(None)
    vectors: Optional[Dict[str, list[float]]] = Field(None)



from weaviate.connect import executor
from weaviate.connect.v4 import ConnectionAsync
from weaviate.debug.executor import _DebugExecutor


@executor.wrap("async")
class _DebugAsync(_DebugExecutor[ConnectionAsync]):
    pass



import httpx


def test_sync_in_sync() -> None:
    res = httpx.get("http://localhost:8000/sync-in-sync")
    assert res.status_code == 200
    assert len(res.json()) == 100


def test_sync_in_async() -> None:
    res = httpx.get("http://localhost:8000/sync-in-async")
    assert res.status_code == 200
    assert len(res.json()) == 100


def test_async_in_async() -> None:
    res = httpx.get("http://localhost:8000/async-in-async")
    assert res.status_code == 200
    assert len(res.json()) == 100



import asyncio
import time
from typing import Awaitable, Callable, cast

from grpc import Call, RpcError, StatusCode  # type: ignore
from grpc.aio import AioRpcError  # type: ignore
from typing_extensions import ParamSpec, TypeVar

from weaviate.exceptions import WeaviateRetryError
from weaviate.logger import logger

P = ParamSpec("P")
T = TypeVar("T")


class _Retry:
    def __init__(self, n: float = 4) -> None:
        self.n = n

    async def awith_exponential_backoff(
        self,
        count: int,
        error: str,
        f: Callable[P, Awaitable[T]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> T:
        try:
            return await f(*args, **kwargs)
        except AioRpcError as e:
            if e.code() != StatusCode.UNAVAILABLE:
                raise e
            logger.info(
                f"{error} received exception: {e}. Retrying with exponential backoff in {2**count} seconds"
            )
            await asyncio.sleep(2**count)
            if count > self.n:
                raise WeaviateRetryError(str(e), count) from e
            return await self.awith_exponential_backoff(count + 1, error, f, *args, **kwargs)

    def with_exponential_backoff(
        self,
        count: int,
        error: str,
        f: Callable[P, T],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> T:
        try:
            return f(*args, **kwargs)
        except RpcError as e:
            err = cast(Call, e)
            if err.code() != StatusCode.UNAVAILABLE:
                raise e
            logger.info(
                f"{error} received exception: {e}. Retrying with exponential backoff in {2**count} seconds"
            )
            time.sleep(2**count)
            if count > self.n:
                raise WeaviateRetryError(str(e), count) from e
            return self.with_exponential_backoff(count + 1, error, f, *args, **kwargs)



from typing import Dict, Optional

from weaviate.aliases.alias import AliasReturn
from weaviate.connect.v4 import ConnectionAsync

from .executor import _AliasExecutor

class _AliasAsync(_AliasExecutor[ConnectionAsync]):
    async def list_all(self, *, collection: Optional[str] = None) -> Dict[str, AliasReturn]: ...
    async def get(self, *, alias_name: str) -> Optional[AliasReturn]: ...
    async def create(self, *, alias_name: str, target_collection: str) -> None: ...
    async def delete(self, *, alias_name: str) -> bool: ...
    async def update(self, *, alias_name: str, new_target_collection: str) -> bool: ...
    async def exists(self, *, alias_name: str) -> bool: ...



from weaviate.collections.aggregations.hybrid.executor import _HybridExecutor
from weaviate.connect import executor
from weaviate.connect.v4 import ConnectionSync


@executor.wrap("sync")
class _Hybrid(_HybridExecutor[ConnectionSync]):
    pass



from .generate import _FetchObjectsByIDsGenerate, _FetchObjectsByIDsGenerateAsync
from .query import _FetchObjectsByIDsQuery, _FetchObjectsByIDsQueryAsync

__all__ = [
    "_FetchObjectsByIDsGenerate",
    "_FetchObjectsByIDsGenerateAsync",
    "_FetchObjectsByIDsQuery",
    "_FetchObjectsByIDsQueryAsync",
]



from weaviate.cluster.replicate.executor import _ReplicateExecutor
from weaviate.connect import executor
from weaviate.connect.v4 import ConnectionAsync


@executor.wrap("async")
class _ReplicateAsync(_ReplicateExecutor[ConnectionAsync]):
    pass



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from weaviate.proto.v1.v6300.v1 import file_replication_pb2 as v1_dot_file__replication__pb2

GRPC_GENERATED_VERSION = '1.72.1'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in v1/file_replication_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class FileReplicationServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.PauseFileActivity = channel.unary_unary(
                '/weaviate.v1.FileReplicationService/PauseFileActivity',
                request_serializer=v1_dot_file__replication__pb2.PauseFileActivityRequest.SerializeToString,
                response_deserializer=v1_dot_file__replication__pb2.PauseFileActivityResponse.FromString,
                _registered_method=True)
        self.ResumeFileActivity = channel.unary_unary(
                '/weaviate.v1.FileReplicationService/ResumeFileActivity',
                request_serializer=v1_dot_file__replication__pb2.ResumeFileActivityRequest.SerializeToString,
                response_deserializer=v1_dot_file__replication__pb2.ResumeFileActivityResponse.FromString,
                _registered_method=True)
        self.ListFiles = channel.unary_unary(
                '/weaviate.v1.FileReplicationService/ListFiles',
                request_serializer=v1_dot_file__replication__pb2.ListFilesRequest.SerializeToString,
                response_deserializer=v1_dot_file__replication__pb2.ListFilesResponse.FromString,
                _registered_method=True)
        self.GetFileMetadata = channel.stream_stream(
                '/weaviate.v1.FileReplicationService/GetFileMetadata',
                request_serializer=v1_dot_file__replication__pb2.GetFileMetadataRequest.SerializeToString,
                response_deserializer=v1_dot_file__replication__pb2.FileMetadata.FromString,
                _registered_method=True)
        self.GetFile = channel.stream_stream(
                '/weaviate.v1.FileReplicationService/GetFile',
                request_serializer=v1_dot_file__replication__pb2.GetFileRequest.SerializeToString,
                response_deserializer=v1_dot_file__replication__pb2.FileChunk.FromString,
                _registered_method=True)


class FileReplicationServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def PauseFileActivity(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ResumeFileActivity(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListFiles(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetFileMetadata(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetFile(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_FileReplicationServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'PauseFileActivity': grpc.unary_unary_rpc_method_handler(
                    servicer.PauseFileActivity,
                    request_deserializer=v1_dot_file__replication__pb2.PauseFileActivityRequest.FromString,
                    response_serializer=v1_dot_file__replication__pb2.PauseFileActivityResponse.SerializeToString,
            ),
            'ResumeFileActivity': grpc.unary_unary_rpc_method_handler(
                    servicer.ResumeFileActivity,
                    request_deserializer=v1_dot_file__replication__pb2.ResumeFileActivityRequest.FromString,
                    response_serializer=v1_dot_file__replication__pb2.ResumeFileActivityResponse.SerializeToString,
            ),
            'ListFiles': grpc.unary_unary_rpc_method_handler(
                    servicer.ListFiles,
                    request_deserializer=v1_dot_file__replication__pb2.ListFilesRequest.FromString,
                    response_serializer=v1_dot_file__replication__pb2.ListFilesResponse.SerializeToString,
            ),
            'GetFileMetadata': grpc.stream_stream_rpc_method_handler(
                    servicer.GetFileMetadata,
                    request_deserializer=v1_dot_file__replication__pb2.GetFileMetadataRequest.FromString,
                    response_serializer=v1_dot_file__replication__pb2.FileMetadata.SerializeToString,
            ),
            'GetFile': grpc.stream_stream_rpc_method_handler(
                    servicer.GetFile,
                    request_deserializer=v1_dot_file__replication__pb2.GetFileRequest.FromString,
                    response_serializer=v1_dot_file__replication__pb2.FileChunk.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'weaviate.v1.FileReplicationService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('weaviate.v1.FileReplicationService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class FileReplicationService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def PauseFileActivity(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/weaviate.v1.FileReplicationService/PauseFileActivity',
            v1_dot_file__replication__pb2.PauseFileActivityRequest.SerializeToString,
            v1_dot_file__replication__pb2.PauseFileActivityResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ResumeFileActivity(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/weaviate.v1.FileReplicationService/ResumeFileActivity',
            v1_dot_file__replication__pb2.ResumeFileActivityRequest.SerializeToString,
            v1_dot_file__replication__pb2.ResumeFileActivityResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ListFiles(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/weaviate.v1.FileReplicationService/ListFiles',
            v1_dot_file__replication__pb2.ListFilesRequest.SerializeToString,
            v1_dot_file__replication__pb2.ListFilesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetFileMetadata(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            '/weaviate.v1.FileReplicationService/GetFileMetadata',
            v1_dot_file__replication__pb2.GetFileMetadataRequest.SerializeToString,
            v1_dot_file__replication__pb2.FileMetadata.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetFile(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            '/weaviate.v1.FileReplicationService/GetFile',
            v1_dot_file__replication__pb2.GetFileRequest.SerializeToString,
            v1_dot_file__replication__pb2.FileChunk.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



import contextvars
import functools
import math
import os
import threading
import time
import uuid as uuid_package
from abc import ABC
from collections import deque
from concurrent.futures import ThreadPoolExecutor
from copy import copy
from dataclasses import dataclass, field
from typing import Any, Dict, Generic, List, Optional, Set, TypeVar, Union, cast

from httpx import ConnectError
from pydantic import ValidationError
from typing_extensions import TypeAlias

from weaviate.cluster.types import Node
from weaviate.collections.batch.grpc_batch_objects import _BatchGRPC
from weaviate.collections.batch.rest import _BatchREST
from weaviate.collections.classes.batch import (
    BatchObject,
    BatchObjectReturn,
    BatchReference,
    BatchReferenceReturn,
    BatchResult,
    ErrorObject,
    ErrorReference,
    Shard,
    _BatchObject,
    _BatchReference,
)
from weaviate.collections.classes.config import ConsistencyLevel
from weaviate.collections.classes.internal import (
    ReferenceInput,
    ReferenceInputs,
    ReferenceToMulti,
)
from weaviate.collections.classes.types import WeaviateProperties
from weaviate.connect import executor
from weaviate.connect.v4 import ConnectionSync
from weaviate.exceptions import EmptyResponseException, WeaviateBatchValidationError
from weaviate.logger import logger
from weaviate.types import UUID, VECTORS
from weaviate.util import _decode_json_response_dict
from weaviate.warnings import _Warnings

BatchResponse = List[Dict[str, Any]]


TBatchInput = TypeVar("TBatchInput")
TBatchReturn = TypeVar("TBatchReturn")
MAX_CONCURRENT_REQUESTS = 10
DEFAULT_REQUEST_TIMEOUT = 180
CONCURRENT_REQUESTS_DYNAMIC_VECTORIZER = 2
BATCH_TIME_TARGET = 10
VECTORIZER_BATCHING_STEP_SIZE = 48  # cohere max batch size is 96
MAX_RETRIES = float(
    os.getenv("WEAVIATE_BATCH_MAX_RETRIES", "9.299")
)  # approximately 10m30s of waiting in worst case, e.g. server scale up event


class BatchRequest(ABC, Generic[TBatchInput, TBatchReturn]):
    """`BatchRequest` abstract class used as a interface for batch requests."""

    def __init__(self) -> None:
        self._items: List[TBatchInput] = []
        self._lock = threading.Lock()

    def __len__(self) -> int:
        return len(self._items)

    def add(self, item: TBatchInput) -> None:
        """Add an item to the BatchRequest."""
        self._lock.acquire()
        self._items.append(item)
        self._lock.release()

    def prepend(self, item: List[TBatchInput]) -> None:
        """Add items to the front of the BatchRequest.

        This is intended to be used when objects should be retries, eg. after a temporary error.
        """
        self._lock.acquire()
        self._items = item + self._items
        self._lock.release()


class ReferencesBatchRequest(BatchRequest[_BatchReference, BatchReferenceReturn]):
    """Collect Weaviate-object references to add them in one request to Weaviate."""

    def pop_items(self, pop_amount: int, uuid_lookup: Set[str]) -> List[_BatchReference]:
        """Pop the given number of items from the BatchRequest queue.

        Returns:
            A list of items from the BatchRequest.
        """
        ret: List[_BatchReference] = []
        i = 0
        self._lock.acquire()
        while len(ret) < pop_amount and len(self._items) > 0 and i < len(self._items):
            if self._items[i].from_uuid not in uuid_lookup and (
                self._items[i].to_uuid is None or self._items[i].to_uuid not in uuid_lookup
            ):
                ret.append(self._items.pop(i))
            else:
                i += 1
        self._lock.release()
        return ret


class ObjectsBatchRequest(BatchRequest[_BatchObject, BatchObjectReturn]):
    """Collect objects for one batch request to weaviate."""

    def pop_items(self, pop_amount: int) -> List[_BatchObject]:
        """Pop the given number of items from the BatchRequest queue.

        Returns:
            A list of items from the BatchRequest.
        """
        self._lock.acquire()
        if pop_amount >= len(self._items):
            ret = copy(self._items)
            self._items.clear()
        else:
            ret = copy(self._items[:pop_amount])
            self._items = self._items[pop_amount:]

        self._lock.release()
        return ret


@dataclass
class _BatchDataWrapper:
    results: BatchResult = field(default_factory=BatchResult)
    failed_objects: List[ErrorObject] = field(default_factory=list)
    failed_references: List[ErrorReference] = field(default_factory=list)
    imported_shards: Set[Shard] = field(default_factory=set)


@dataclass
class _DynamicBatching:
    pass


@dataclass
class _FixedSizeBatching:
    batch_size: int
    concurrent_requests: int


@dataclass
class _RateLimitedBatching:
    requests_per_minute: int


_BatchMode: TypeAlias = Union[_DynamicBatching, _FixedSizeBatching, _RateLimitedBatching]


class _BatchBase:
    def __init__(
        self,
        connection: ConnectionSync,
        consistency_level: Optional[ConsistencyLevel],
        results: _BatchDataWrapper,
        batch_mode: _BatchMode,
        executor: ThreadPoolExecutor,
        vectorizer_batching: bool,
        objects: Optional[ObjectsBatchRequest] = None,
        references: Optional[ReferencesBatchRequest] = None,
    ) -> None:
        self.__batch_objects = objects or ObjectsBatchRequest()
        self.__batch_references = references or ReferencesBatchRequest()

        self.__connection = connection
        self.__consistency_level: Optional[ConsistencyLevel] = consistency_level
        self.__vectorizer_batching = vectorizer_batching

        self.__batch_grpc = _BatchGRPC(connection._weaviate_version, self.__consistency_level)
        self.__batch_rest = _BatchREST(self.__consistency_level)

        # lookup table for objects that are currently being processed - is used to not send references from objects that have not been added yet
        self.__uuid_lookup: Set[str] = set()

        # we do not want that users can access the results directly as they are not thread-safe
        self.__results_for_wrapper_backup = results
        self.__results_for_wrapper = _BatchDataWrapper()

        self.__cluster = _ClusterBatch(self.__connection)

        self.__batching_mode: _BatchMode = batch_mode
        self.__max_batch_size: int = 1000

        self.__executor = executor
        self.__objs_count = 0
        self.__objs_logs_count = 0
        self.__refs_logs_count = 0

        if isinstance(self.__batching_mode, _FixedSizeBatching):
            self.__recommended_num_objects = self.__batching_mode.batch_size
            self.__concurrent_requests = self.__batching_mode.concurrent_requests
        elif isinstance(self.__batching_mode, _RateLimitedBatching):
            # Batch with rate limiting should never send more than the given amount of objects per minute.
            # We could send all objects in a single batch every 60 seconds but that could cause problems with too large requests. Therefore, we
            # limit the size of a batch to self.__max_batch_size and send multiple batches of equal size and send them in equally space in time.
            # Example:
            #  3000 objects, 1000/min -> 3 batches of 1000 objects, send every 20 seconds
            self.__concurrent_requests = (
                self.__batching_mode.requests_per_minute + self.__max_batch_size
            ) // self.__max_batch_size
            self.__recommended_num_objects = (
                self.__batching_mode.requests_per_minute // self.__concurrent_requests
            )
        elif isinstance(self.__batching_mode, _DynamicBatching) and not self.__vectorizer_batching:
            self.__recommended_num_objects = 10
            self.__concurrent_requests = 2
        else:
            assert isinstance(self.__batching_mode, _DynamicBatching) and self.__vectorizer_batching
            self.__recommended_num_objects = VECTORIZER_BATCHING_STEP_SIZE
            self.__concurrent_requests = 2
            self.__dynamic_batching_sleep_time: int = 0
            self._batch_send: bool = False

        self.__recommended_num_refs: int = 50

        self.__active_requests = 0

        # dynamic batching
        self.__time_last_scale_up: float = 0
        self.__rate_queue: deque = deque(maxlen=50)  # 5s with 0.1s refresh rate
        self.__took_queue: deque = deque(maxlen=CONCURRENT_REQUESTS_DYNAMIC_VECTORIZER)

        # fixed rate batching
        self.__time_stamp_last_request: float = 0
        # do 62 secs to give us some buffer to the "per-minute" calculation
        self.__fix_rate_batching_base_time = 62

        self.__active_requests_lock = threading.Lock()
        self.__uuid_lookup_lock = threading.Lock()
        self.__results_lock = threading.Lock()

        self.__bg_thread = self.__start_bg_threads()
        self.__bg_thread_exception: Optional[Exception] = None

    @property
    def number_errors(self) -> int:
        """Return the number of errors in the batch."""
        return len(self.__results_for_wrapper.failed_objects) + len(
            self.__results_for_wrapper.failed_references
        )

    def _shutdown(self) -> None:
        """Shutdown the current batch and wait for all requests to be finished."""
        self.flush()

        # we are done, shut bg threads down and end the event loop
        self.__shut_background_thread_down.set()
        while self.__bg_thread.is_alive():
            time.sleep(0.01)

        # copy the results to the public results
        self.__results_for_wrapper_backup.results = self.__results_for_wrapper.results
        self.__results_for_wrapper_backup.failed_objects = self.__results_for_wrapper.failed_objects
        self.__results_for_wrapper_backup.failed_references = (
            self.__results_for_wrapper.failed_references
        )
        self.__results_for_wrapper_backup.imported_shards = (
            self.__results_for_wrapper.imported_shards
        )

    def __batch_send(self) -> None:
        refresh_time: float = 0.01
        while (
            self.__shut_background_thread_down is not None
            and not self.__shut_background_thread_down.is_set()
        ):
            if isinstance(self.__batching_mode, _RateLimitedBatching):
                if (
                    time.time() - self.__time_stamp_last_request
                    < self.__fix_rate_batching_base_time // self.__concurrent_requests
                ):
                    time.sleep(1)
                    continue
                refresh_time = 0
            elif isinstance(self.__batching_mode, _DynamicBatching) and self.__vectorizer_batching:
                if self.__dynamic_batching_sleep_time > 0:
                    if (
                        time.time() - self.__time_stamp_last_request
                        < self.__dynamic_batching_sleep_time
                    ):
                        time.sleep(1)
                        continue

            if (
                self.__active_requests < self.__concurrent_requests
                and len(self.__batch_objects) + len(self.__batch_references) > 0
            ):
                self.__time_stamp_last_request = time.time()

                self._batch_send = True
                with self.__active_requests_lock:
                    self.__active_requests += 1

                start = time.time()
                while (len_o := len(self.__batch_objects)) < self.__recommended_num_objects and (
                    len_r := len(self.__batch_references)
                ) < self.__recommended_num_refs:
                    # wait for more objects to be added up to the recommended number
                    time.sleep(0.01)
                    if (
                        self.__shut_background_thread_down is not None
                        and self.__shut_background_thread_down.is_set()
                    ):
                        # shutdown was requested, exit the loop
                        break
                    if time.time() - start >= 1 and (
                        len_o == len(self.__batch_objects) or len_r == len(self.__batch_references)
                    ):
                        # no new objects were added in the last second, exit the loop
                        break

                objs = self.__batch_objects.pop_items(self.__recommended_num_objects)
                refs = self.__batch_references.pop_items(
                    self.__recommended_num_refs,
                    uuid_lookup=self.__uuid_lookup,
                )
                # do not block the thread - the results are written to a central (locked) list and we want to have multiple concurrent batch-requests
                ctx = contextvars.copy_context()
                self.__executor.submit(
                    ctx.run,
                    functools.partial(
                        self.__send_batch,
                        objs,
                        refs,
                        readd_rate_limit=isinstance(self.__batching_mode, _RateLimitedBatching),
                    ),
                )

            time.sleep(refresh_time)

    def __dynamic_batch_rate_loop(self) -> None:
        refresh_time = 1
        while (
            self.__shut_background_thread_down is not None
            and not self.__shut_background_thread_down.is_set()
        ):
            if not isinstance(self.__batching_mode, _DynamicBatching):
                return

            try:
                self.__dynamic_batching()
            except Exception as e:
                logger.debug(repr(e))

            time.sleep(refresh_time)

    def __start_bg_threads(self) -> threading.Thread:
        """Create a background thread that periodically checks how congested the batch queue is."""
        self.__shut_background_thread_down = threading.Event()

        def dynamic_batch_rate_wrapper() -> None:
            try:
                self.__dynamic_batch_rate_loop()
            except Exception as e:
                self.__bg_thread_exception = e

        demonDynamic = threading.Thread(
            target=dynamic_batch_rate_wrapper,
            daemon=True,
            name="BgDynamicBatchRate",
        )
        demonDynamic.start()

        def batch_send_wrapper() -> None:
            try:
                self.__batch_send()
            except Exception as e:
                logger.error(e)
                self.__bg_thread_exception = e

        demonBatchSend = threading.Thread(
            target=batch_send_wrapper,
            daemon=True,
            name="BgBatchScheduler",
        )
        demonBatchSend.start()

        return demonBatchSend

    def __dynamic_batching(self) -> None:
        status = self.__cluster.get_nodes_status()
        if "batchStats" not in status[0] or "queueLength" not in status[0]["batchStats"]:
            # async indexing - just send a lot
            self.__batching_mode = _FixedSizeBatching(1000, 10)
            self.__recommended_num_objects = 1000
            self.__concurrent_requests = 10
            return

        rate: int = status[0]["batchStats"]["ratePerSecond"]
        rate_per_worker = rate / self.__concurrent_requests

        batch_length = status[0]["batchStats"]["queueLength"]

        self.__rate_queue.append(rate)

        if self.__vectorizer_batching:
            # slow vectorizer, we want to send larger batches that can take a bit longer, but fewer of them. We might need to sleep
            if len(self.__took_queue) > 0 and self._batch_send:
                max_took = max(self.__took_queue)
                self.__dynamic_batching_sleep_time = 0
                if max_took > 2 * BATCH_TIME_TARGET:
                    self.__concurrent_requests = 1
                    self.__recommended_num_objects = VECTORIZER_BATCHING_STEP_SIZE
                elif max_took > BATCH_TIME_TARGET:
                    current_step = self.__recommended_num_objects // VECTORIZER_BATCHING_STEP_SIZE

                    if self.__concurrent_requests > 1:
                        self.__concurrent_requests -= 1
                    elif current_step > 1:
                        self.__recommended_num_objects = VECTORIZER_BATCHING_STEP_SIZE * (
                            current_step - 1
                        )
                    else:
                        # cannot scale down, sleep a bit
                        self.__dynamic_batching_sleep_time = max_took - BATCH_TIME_TARGET

                elif max_took < 3 * BATCH_TIME_TARGET // 4:
                    if self.__dynamic_batching_sleep_time > 0:
                        self.__dynamic_batching_sleep_time = 0
                    elif self.__concurrent_requests < 3:
                        self.__concurrent_requests += 1
                    else:
                        current_step = (
                            self.__recommended_num_objects // VECTORIZER_BATCHING_STEP_SIZE
                        )
                        self.__recommended_num_objects = VECTORIZER_BATCHING_STEP_SIZE * (
                            current_step + 1
                        )
                self._batch_send = False
        else:
            if batch_length == 0:  # scale up if queue is empty
                self.__recommended_num_objects = min(
                    self.__recommended_num_objects + 50,
                    self.__max_batch_size,
                )

                if (
                    self.__max_batch_size == self.__recommended_num_objects
                    and len(self.__batch_objects) > self.__recommended_num_objects
                    and time.time() - self.__time_last_scale_up > 1
                    and self.__concurrent_requests < MAX_CONCURRENT_REQUESTS
                ):
                    self.__concurrent_requests += 1
                    self.__time_last_scale_up = time.time()

            else:
                ratio = batch_length / rate
                if 2.1 > ratio > 1.9:  # ideal, send exactly as many objects as weaviate can process
                    self.__recommended_num_objects = math.floor(rate_per_worker)
                elif ratio <= 1.9:  # we can send more
                    self.__recommended_num_objects = math.floor(
                        min(
                            self.__recommended_num_objects * 1.5,
                            rate_per_worker * 2 / ratio,
                        )
                    )

                    if self.__max_batch_size == self.__recommended_num_objects:
                        self.__concurrent_requests += 1

                elif ratio < 10:  # too high, scale down
                    self.__recommended_num_objects = math.floor(rate_per_worker * 2 / ratio)

                    if self.__recommended_num_objects < 100 and self.__concurrent_requests > 2:
                        self.__concurrent_requests -= 1

                else:  # way too high, stop sending new batches
                    self.__recommended_num_objects = 0
                    self.__concurrent_requests = 2

    def __send_batch(
        self,
        objs: List[_BatchObject],
        refs: List[_BatchReference],
        readd_rate_limit: bool,
    ) -> None:
        if (n_objs := len(objs)) > 0:
            start = time.time()
            try:
                response_obj = executor.result(
                    self.__batch_grpc.objects(
                        connection=self.__connection,
                        objects=objs,
                        timeout=DEFAULT_REQUEST_TIMEOUT,
                        max_retries=MAX_RETRIES,
                    )
                )
                if response_obj.has_errors:
                    logger.error(
                        {
                            "message": f"Failed to send {len(response_obj.errors)} in a batch of {len(objs)}",
                            "errors": {err.message for err in response_obj.errors.values()},
                        }
                    )
            except Exception as e:
                errors_obj = {
                    idx: ErrorObject(message=repr(e), object_=BatchObject._from_internal(obj))
                    for idx, obj in enumerate(objs)
                }
                logger.error(
                    {
                        "message": f"Failed to send all objects in a batch of {len(objs)}",
                        "error": repr(e),
                    }
                )
                response_obj = BatchObjectReturn(
                    _all_responses=list(errors_obj.values()),
                    elapsed_seconds=time.time() - start,
                    errors=errors_obj,
                    has_errors=True,
                )

            readded_uuids = set()
            readded_objects = []
            highest_retry_count = 0
            for i, err in response_obj.errors.items():
                if (
                    (
                        "support@cohere.com" in err.message
                        and (
                            "rate limit" in err.message
                            or "500 error: internal server error" in err.message
                        )
                    )
                    or (
                        "OpenAI" in err.message
                        and (
                            "Rate limit reached" in err.message
                            or "on tokens per min (TPM)" in err.message
                            or "503 error: Service Unavailable." in err.message
                            or "500 error: The server had an error while processing your request."
                            in err.message
                        )
                    )
                    or ("failed with status: 503 error" in err.message)  # huggingface
                ):
                    if err.object_.retry_count > highest_retry_count:
                        highest_retry_count = err.object_.retry_count

                    if err.object_.retry_count > 5:
                        continue  # too many retries, give up
                    err.object_.retry_count += 1
                    readded_objects.append(i)

            if len(readded_objects) > 0:
                _Warnings.batch_rate_limit_reached(
                    response_obj.errors[readded_objects[0]].message,
                    self.__fix_rate_batching_base_time * (highest_retry_count + 1),
                )

                readd_objects = [
                    err.object_._to_internal()
                    for i, err in response_obj.errors.items()
                    if i in readded_objects
                ]
                readded_uuids = {obj.uuid for obj in readd_objects}

                self.__batch_objects.prepend(readd_objects)

                new_errors = {
                    i: err for i, err in response_obj.errors.items() if i not in readded_objects
                }
                response_obj = BatchObjectReturn(
                    uuids={
                        i: uid for i, uid in response_obj.uuids.items() if i not in readded_objects
                    },
                    errors=new_errors,
                    has_errors=len(new_errors) > 0,
                    _all_responses=[
                        err
                        for i, err in enumerate(response_obj.all_responses)
                        if i not in readded_objects
                    ],
                    elapsed_seconds=response_obj.elapsed_seconds,
                )
                if readd_rate_limit:
                    # for rate limited batching the timing is handled by the outer loop => no sleep here
                    self.__time_stamp_last_request = (
                        time.time() + self.__fix_rate_batching_base_time * (highest_retry_count + 1)
                    )  # skip a full minute to recover from the rate limit
                    self.__fix_rate_batching_base_time += (
                        1  # increase the base time as the current one is too low
                    )
                else:
                    # sleep a bit to recover from the rate limit in other cases
                    time.sleep(2**highest_retry_count)
            with self.__uuid_lookup_lock:
                self.__uuid_lookup.difference_update(
                    obj.uuid for obj in objs if obj.uuid not in readded_uuids
                )

            if (n_obj_errs := len(response_obj.errors)) > 0 and self.__objs_logs_count < 30:
                logger.error(
                    {
                        "message": f"Failed to send {n_obj_errs} objects in a batch of {n_objs}. Please inspect client.batch.failed_objects or collection.batch.failed_objects for the failed objects.",
                    }
                )
                self.__objs_logs_count += 1
            if self.__objs_logs_count > 30:
                logger.error(
                    {
                        "message": "There have been more than 30 failed object batches. Further errors will not be logged.",
                    }
                )
            with self.__results_lock:
                self.__results_for_wrapper.results.objs += response_obj
                self.__results_for_wrapper.failed_objects.extend(response_obj.errors.values())
            self.__took_queue.append(time.time() - start)

        if (n_refs := len(refs)) > 0:
            start = time.time()
            try:
                response_ref = executor.result(
                    self.__batch_rest.references(connection=self.__connection, references=refs)
                )
            except Exception as e:
                errors_ref = {
                    idx: ErrorReference(
                        message=repr(e), reference=BatchReference._from_internal(ref)
                    )
                    for idx, ref in enumerate(refs)
                }
                response_ref = BatchReferenceReturn(
                    elapsed_seconds=time.time() - start,
                    errors=errors_ref,
                    has_errors=True,
                )
            if (n_ref_errs := len(response_ref.errors)) > 0 and self.__refs_logs_count < 30:
                logger.error(
                    {
                        "message": f"Failed to send {n_ref_errs} references in a batch of {n_refs}. Please inspect client.batch.failed_references or collection.batch.failed_references for the failed references.",
                        "errors": response_ref.errors,
                    }
                )
                self.__refs_logs_count += 1
            if self.__refs_logs_count > 30:
                logger.error(
                    {
                        "message": "There have been more than 30 failed reference batches. Further errors will not be logged.",
                    }
                )
            with self.__results_lock:
                self.__results_for_wrapper.results.refs += response_ref
                self.__results_for_wrapper.failed_references.extend(response_ref.errors.values())

        with self.__active_requests_lock:
            self.__active_requests -= 1

    def flush(self) -> None:
        """Flush the batch queue and wait for all requests to be finished."""
        # bg thread is sending objs+refs automatically, so simply wait for everything to be done
        while (
            self.__active_requests > 0
            or len(self.__batch_objects) > 0
            or len(self.__batch_references) > 0
        ):
            time.sleep(0.01)
            self.__check_bg_thread_alive()

    def _add_object(
        self,
        collection: str,
        properties: Optional[WeaviateProperties] = None,
        references: Optional[ReferenceInputs] = None,
        uuid: Optional[UUID] = None,
        vector: Optional[VECTORS] = None,
        tenant: Optional[str] = None,
    ) -> UUID:
        self.__check_bg_thread_alive()
        try:
            batch_object = BatchObject(
                collection=collection,
                properties=properties,
                references=references,
                uuid=uuid,
                vector=vector,
                tenant=tenant,
                index=self.__objs_count,
            )
            self.__objs_count += 1
            self.__results_for_wrapper.imported_shards.add(
                Shard(collection=collection, tenant=tenant)
            )
        except ValidationError as e:
            raise WeaviateBatchValidationError(repr(e))
        self.__uuid_lookup.add(str(batch_object.uuid))
        self.__batch_objects.add(batch_object._to_internal())

        # block if queue gets too long or weaviate is overloaded - reading files is faster them sending them so we do
        # not need a long queue
        while (
            self.__recommended_num_objects == 0
            or len(self.__batch_objects) >= self.__recommended_num_objects * 2
        ):
            self.__check_bg_thread_alive()
            time.sleep(0.01)

        assert batch_object.uuid is not None
        return batch_object.uuid

    def _add_reference(
        self,
        from_object_uuid: UUID,
        from_object_collection: str,
        from_property_name: str,
        to: ReferenceInput,
        tenant: Optional[str] = None,
    ) -> None:
        self.__check_bg_thread_alive()
        if isinstance(to, ReferenceToMulti):
            to_strs: Union[List[str], List[UUID]] = to.uuids_str
        elif isinstance(to, str) or isinstance(to, uuid_package.UUID):
            to_strs = [to]
        else:
            to_strs = list(to)

        for uid in to_strs:
            try:
                batch_reference = BatchReference(
                    from_object_collection=from_object_collection,
                    from_object_uuid=from_object_uuid,
                    from_property_name=from_property_name,
                    to_object_collection=(
                        to.target_collection if isinstance(to, ReferenceToMulti) else None
                    ),
                    to_object_uuid=uid,
                    tenant=tenant,
                )
            except ValidationError as e:
                raise WeaviateBatchValidationError(repr(e))
            self.__batch_references.add(batch_reference._to_internal())

        # block if queue gets too long or weaviate is overloaded
        while self.__recommended_num_objects == 0:
            time.sleep(0.01)  # block if weaviate is overloaded, also do not send any refs
            self.__check_bg_thread_alive()

    def __check_bg_thread_alive(self) -> None:
        if self.__bg_thread.is_alive():
            return

        raise self.__bg_thread_exception or Exception("Batch thread died unexpectedly")


class _ClusterBatch:
    def __init__(self, connection: ConnectionSync):
        self._connection = connection

    def get_nodes_status(
        self,
    ) -> List[Node]:
        try:
            response = executor.result(self._connection.get(path="/nodes"))
        except ConnectError as conn_err:
            raise ConnectError("Get nodes status failed due to connection error") from conn_err

        response_typed = _decode_json_response_dict(response, "Nodes status")
        assert response_typed is not None
        nodes = response_typed.get("nodes")
        if nodes is None or nodes == []:
            raise EmptyResponseException("Nodes status response returned empty")
        return cast(List[Node], nodes)



import pytest
from pydantic import ValidationError

from weaviate.collections.classes.filters import Filter
from weaviate.collections.classes.grpc import (
    QueryReference,
    Sort,
    _QueryReference,
    _QueryReferenceMultiTarget,
)


def test_link_to_errors_on_extra_variable() -> None:
    with pytest.raises(ValidationError):
        QueryReference(link_on="ref", return_property="name")


def test_link_to_multi_target_errors_on_extra_variable() -> None:
    with pytest.raises(ValidationError):
        _QueryReferenceMultiTarget(link_on="ref", target_colection="Test")


def test_link_to_multi_target_is_link_to() -> None:
    link_to = _QueryReferenceMultiTarget(link_on="ref", target_collection="Test")
    assert isinstance(link_to, _QueryReference)


def test_direct_init_filter() -> None:
    with pytest.raises(TypeError):
        Filter()


def test_direct_init_sort() -> None:
    with pytest.raises(TypeError):
        Sort()



from .generate import _NearMediaGenerate, _NearMediaGenerateAsync
from .query import _NearMediaQuery, _NearMediaQueryAsync

__all__ = [
    "_NearMediaGenerate",
    "_NearMediaQuery",
    "_NearMediaGenerateAsync",
    "_NearMediaQueryAsync",
]



# MetaEnum and BaseEnum are required to support `in` statements:
#    'ALL' in ConsistencyLevel == True
#    12345 in ConsistencyLevel == False
from enum import Enum, EnumMeta
from typing import Any


class MetaEnum(EnumMeta):
    def __contains__(cls, item: Any) -> bool:
        try:
            # when item is type ConsistencyLevel
            return item.name in cls.__members__.keys()
        except AttributeError:
            # when item is type str
            return item in cls.__members__.keys()


class BaseEnum(Enum, metaclass=MetaEnum):
    pass



from typing import Generic

from weaviate.collections.classes.types import Properties, References
from weaviate.collections.queries.near_object.query.executor import (
    _NearObjectQueryExecutor,
)
from weaviate.connect import executor
from weaviate.connect.v4 import ConnectionSync


@executor.wrap("sync")
class _NearObjectQuery(
    Generic[Properties, References],
    _NearObjectQueryExecutor[ConnectionSync, Properties, References],
):
    pass



import struct
import uuid as uuid_lib
from dataclasses import dataclass
from typing import (
    Any,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    Union,
    cast,
    get_args,
)

from typing_extensions import TypeGuard

from weaviate.collections.classes.config import ConsistencyLevel
from weaviate.collections.classes.grpc import (
    BM25OperatorOptions,
    BM25OperatorOr,
    HybridFusion,
    HybridVectorType,
    Move,
    NearVectorInputType,
    OneDimensionalVectorType,
    PrimitiveVectorType,
    TargetVectorJoinType,
    TwoDimensionalVectorType,
    _HybridNearText,
    _HybridNearVector,
    _ListOfVectorsQuery,
    _MultiTargetVectorJoin,
)
from weaviate.exceptions import (
    WeaviateInvalidInputError,
)
from weaviate.proto.v1 import base_pb2, base_search_pb2
from weaviate.types import NUMBER, UUID
from weaviate.util import _get_vector_v4, _ServerVersion
from weaviate.validator import (
    _ExtraTypes,
    _is_valid,
    _validate_input,
    _ValidateArgument,
)

UINT32_LEN = 4
UINT64_LEN = 8


class _BaseGRPC:
    def __init__(
        self,
        weaviate_version: _ServerVersion,
        consistency_level: Optional[ConsistencyLevel],
        validate_arguments: bool,
    ):
        self._weaviate_version = weaviate_version
        self._consistency_level = self._get_consistency_level(consistency_level)
        self._validate_arguments = validate_arguments

    @staticmethod
    def _get_consistency_level(
        consistency_level: Optional[ConsistencyLevel],
    ) -> Optional["base_pb2.ConsistencyLevel"]:
        if consistency_level is None:
            return None

        if consistency_level.value == ConsistencyLevel.ONE:
            return base_pb2.ConsistencyLevel.CONSISTENCY_LEVEL_ONE
        elif consistency_level.value == ConsistencyLevel.QUORUM:
            return base_pb2.ConsistencyLevel.CONSISTENCY_LEVEL_QUORUM
        else:
            assert consistency_level.value == ConsistencyLevel.ALL
            return base_pb2.ConsistencyLevel.CONSISTENCY_LEVEL_ALL

    def _recompute_target_vector_to_grpc(
        self,
        target_vector: Optional[TargetVectorJoinType],
        target_vectors_tmp: List[str],
    ) -> Tuple[Optional[base_search_pb2.Targets], Optional[List[str]]]:
        # reorder input for targets so they match the vectors
        if isinstance(target_vector, _MultiTargetVectorJoin):
            target_vector.target_vectors = target_vectors_tmp
            if target_vector.weights is not None:
                target_vector.weights = {
                    target: target_vector.weights[target] for target in target_vectors_tmp
                }
        else:
            target_vector = target_vectors_tmp
        return self.__target_vector_to_grpc(target_vector)

    def __target_vector_to_grpc(
        self, target_vector: Optional[TargetVectorJoinType]
    ) -> Tuple[Optional[base_search_pb2.Targets], Optional[List[str]]]:
        if target_vector is None:
            return None, None

        if isinstance(target_vector, str):
            return base_search_pb2.Targets(target_vectors=[target_vector]), None
        elif isinstance(target_vector, list):
            return base_search_pb2.Targets(target_vectors=target_vector), None
        else:
            return target_vector.to_grpc_target_vector(self._weaviate_version), None

    def _vector_per_target(
        self,
        vector: NearVectorInputType,
        targets: Optional[base_search_pb2.Targets],
        argument_name: str,
    ) -> Tuple[Optional[Dict[str, bytes]], Optional[bytes]]:
        """@deprecated in 1.27.0, included for BC until 1.27.0 is no longer supported."""  # noqa: D401
        invalid_nv_exception = WeaviateInvalidInputError(
            f"""{argument_name} argument can be:
                                - a list of numbers
                                - a dictionary with target names as keys and lists of numbers as values
                        received: {vector}"""
        )
        if isinstance(vector, dict):
            if targets is None or len(targets.target_vectors) != len(vector):
                raise WeaviateInvalidInputError(
                    "The number of target vectors must be equal to the number of vectors."
                )

            vector_per_target: Dict[str, bytes] = {}
            for key, value in vector.items():
                nv = _get_vector_v4(value)

                if (
                    not isinstance(nv, list)
                    or len(nv) == 0
                    or not isinstance(nv[0], get_args(NUMBER))
                ):
                    raise invalid_nv_exception

                vector_per_target[key] = struct.pack("{}f".format(len(nv)), *nv)

            return vector_per_target, None
        else:
            if isinstance(vector, _ListOfVectorsQuery) or len(vector) == 0:
                raise invalid_nv_exception

            if _is_1d_vector(vector):
                near_vector = _get_vector_v4(vector)
                if not isinstance(near_vector, list):
                    raise invalid_nv_exception
                return None, struct.pack("{}f".format(len(near_vector)), *near_vector)
            else:
                raise WeaviateInvalidInputError(
                    """This input appears to be a nested list of embeddings.
                    If you are trying to search with a multi-vector embedding, check the shape of your input.
                    If you are trying to provide multiple target vectors,
                    provide a dictionary with target names as keys and embeddings as values."""
                )

    def _vector_for_target(
        self,
        vector: NearVectorInputType,
        targets: Optional[base_search_pb2.Targets],
        argument_name: str,
    ) -> Tuple[
        Optional[List[base_search_pb2.VectorForTarget]],
        Optional[bytes],
        Optional[List[str]],
    ]:
        invalid_nv_exception = WeaviateInvalidInputError(
            f"""{argument_name} argument can be:
                                - a list of numbers
                                - a dictionary with target names as keys and lists of numbers as values for multi target search. The keys must match the given target vectors
                        received: {vector} and {targets}."""
        )

        vector_for_target: List[base_search_pb2.VectorForTarget] = []
        target_vectors: List[str] = []

        def add_1d_vector(val: OneDimensionalVectorType, key: str) -> None:
            vec = _get_vector_v4(val)

            if (
                not isinstance(vec, list)
                or len(vec) == 0
                or not isinstance(vec[0], get_args(NUMBER))
            ):
                raise invalid_nv_exception

            if self._weaviate_version.is_lower_than(1, 29, 0):
                vector_for_target.append(
                    base_search_pb2.VectorForTarget(name=key, vector_bytes=_Pack.single(vec))
                )
            else:
                vector_for_target.append(
                    base_search_pb2.VectorForTarget(
                        name=key,
                        vectors=[
                            base_pb2.Vectors(
                                name=key,
                                vector_bytes=_Pack.single(vec),
                                type=base_pb2.Vectors.VECTOR_TYPE_SINGLE_FP32,
                            )
                        ],
                    )
                )
            target_vectors.append(key)

        def add_2d_vector(value: TwoDimensionalVectorType, key: str) -> None:
            if self._weaviate_version.is_lower_than(1, 29, 0):
                for v in value:
                    add_1d_vector(v, key)
                return
            vector_for_target.append(
                base_search_pb2.VectorForTarget(
                    name=key,
                    vectors=[
                        base_pb2.Vectors(
                            name=key,
                            vector_bytes=_Pack.multi([_get_vector_v4(v) for v in value]),
                            type=base_pb2.Vectors.VECTOR_TYPE_MULTI_FP32,
                        )
                    ],
                )
            )
            target_vectors.append(key)

        def add_list_of_vectors(value: _ListOfVectorsQuery, key: str) -> None:
            if _ListOfVectorsQuery.is_one_dimensional(
                value
            ) and self._weaviate_version.is_lower_than(1, 29, 0):
                for v in value.vectors:
                    add_1d_vector(v, key)
                return
            elif _ListOfVectorsQuery.is_one_dimensional(
                value
            ) and self._weaviate_version.is_at_least(1, 29, 0):
                vectors = [
                    base_pb2.Vectors(
                        name=key,
                        vector_bytes=_Pack.multi([_get_vector_v4(v) for v in value.vectors]),
                        type=base_pb2.Vectors.VECTOR_TYPE_MULTI_FP32,
                    )
                ]
            elif _ListOfVectorsQuery.is_two_dimensional(value):
                vectors = [
                    base_pb2.Vectors(
                        name=key,
                        vector_bytes=_Pack.multi([_get_vector_v4(v) for v in vecs]),
                        type=base_pb2.Vectors.VECTOR_TYPE_MULTI_FP32,
                    )
                    for vecs in value.vectors
                ]
            else:
                raise WeaviateInvalidInputError(f"Invalid list of vectors: {value}")
            vector_for_target.append(
                base_search_pb2.VectorForTarget(
                    name=key,
                    vectors=vectors,
                )
            )
            target_vectors.append(key)

        if isinstance(vector, dict):
            if (
                len(vector) == 0
                or targets is None
                or len(set(targets.target_vectors)) != len(vector)
            ):
                raise invalid_nv_exception
            for key, value in vector.items():
                if _is_1d_vector(value):
                    add_1d_vector(value, key)
                elif _is_2d_vector(value):
                    add_2d_vector(value, key)
                elif isinstance(value, _ListOfVectorsQuery):
                    add_list_of_vectors(value, key)
                else:
                    raise invalid_nv_exception
            return vector_for_target, None, target_vectors
        else:
            if _is_1d_vector(vector):
                near_vector = _get_vector_v4(vector)
                if not isinstance(near_vector, list):
                    raise invalid_nv_exception
                return (
                    None,
                    struct.pack("{}f".format(len(near_vector)), *near_vector),
                    None,
                )
            else:
                raise WeaviateInvalidInputError(
                    """This input appears to be a nested list of embeddings.
                    If you are trying to search with a multi-vector embedding, check the shape of your input.
                    If you are trying to provide multiple target vectors,
                    provide a dictionary with target names as keys and embeddings as values."""
                )

    def _parse_near_options(
        self,
        certainty: Optional[NUMBER] = None,
        distance: Optional[NUMBER] = None,
    ) -> Tuple[Optional[float], Optional[float]]:
        if self._validate_arguments:
            _validate_input(
                [
                    _ValidateArgument([float, int, None], "certainty", certainty),
                    _ValidateArgument([float, int, None], "distance", distance),
                ]
            )
        return (
            float(certainty) if certainty is not None else None,
            float(distance) if distance is not None else None,
        )

    def _parse_near_vector(
        self,
        near_vector: NearVectorInputType,
        certainty: Optional[NUMBER],
        distance: Optional[NUMBER],
        target_vector: Optional[TargetVectorJoinType],
    ) -> base_search_pb2.NearVector:
        if self._validate_arguments:
            _validate_input(
                [
                    _ValidateArgument(
                        [
                            List,
                            Dict,
                            _ExtraTypes.PANDAS,
                            _ExtraTypes.POLARS,
                            _ExtraTypes.NUMPY,
                            _ExtraTypes.TF,
                        ],
                        "near_vector",
                        near_vector,
                    ),
                    _ValidateArgument(
                        [str, None, List, _MultiTargetVectorJoin],
                        "target_vector",
                        target_vector,
                    ),
                ]
            )

        certainty, distance = self._parse_near_options(certainty, distance)

        targets, target_vectors = self.__target_vector_to_grpc(target_vector)

        if _is_1d_vector(near_vector) and len(near_vector) > 0:
            # fast path for simple single-vector
            if self._weaviate_version.is_lower_than(1, 29, 0):
                near_vector_grpc: Optional[bytes] = struct.pack(
                    "{}f".format(len(near_vector)), *near_vector
                )
                vector_per_target_tmp = None
                vector_for_targets = None
                vectors = None
            else:
                near_vector_grpc = None
                vector_per_target_tmp = None
                vector_for_targets = None
                vectors = [
                    base_pb2.Vectors(
                        vector_bytes=_Pack.single(near_vector),
                        type=base_pb2.Vectors.VECTOR_TYPE_SINGLE_FP32,
                    )
                ]
        elif _is_2d_vector(near_vector) and self._weaviate_version.is_at_least(1, 29, 0):
            # fast path for simple multi-vector
            near_vector_grpc = None
            vector_per_target_tmp = None
            vector_for_targets = None
            vectors = [
                base_pb2.Vectors(
                    vector_bytes=_Pack.multi(near_vector),
                    type=base_pb2.Vectors.VECTOR_TYPE_MULTI_FP32,
                )
            ]
        else:
            if self._weaviate_version.is_lower_than(1, 27, 0):
                vector_per_target_tmp, near_vector_grpc = self._vector_per_target(
                    near_vector, targets, "near_vector"
                )
                vector_for_targets = None
            else:
                vector_for_targets, near_vector_grpc, target_vectors_tmp = self._vector_for_target(
                    near_vector, targets, "near_vector"
                )
                vector_per_target_tmp = None
                if target_vectors_tmp is not None:
                    targets, target_vectors = self._recompute_target_vector_to_grpc(
                        target_vector, target_vectors_tmp
                    )
            vectors = None
        return base_search_pb2.NearVector(
            vector_bytes=near_vector_grpc,
            certainty=certainty,
            distance=distance,
            targets=targets,
            target_vectors=target_vectors,
            vector_per_target=vector_per_target_tmp,
            vector_for_targets=vector_for_targets,
            vectors=vectors,
        )

    @staticmethod
    def __parse_move(
        move: Optional[Move],
    ) -> Optional[base_search_pb2.NearTextSearch.Move]:
        return (
            base_search_pb2.NearTextSearch.Move(
                force=move.force,
                concepts=move._concepts_list,
                uuids=move._objects_list,
            )
            if move is not None
            else None
        )

    def _parse_near_text(
        self,
        near_text: Union[List[str], str],
        certainty: Optional[NUMBER],
        distance: Optional[NUMBER],
        move_to: Optional[Move],
        move_away: Optional[Move],
        target_vector: Optional[TargetVectorJoinType],
    ) -> base_search_pb2.NearTextSearch:
        if self._validate_arguments:
            _validate_input(
                [
                    _ValidateArgument([List, str], "near_text", near_text),
                    _ValidateArgument([Move, None], "move_away", move_away),
                    _ValidateArgument([Move, None], "move_to", move_to),
                    _ValidateArgument(
                        [str, List, _MultiTargetVectorJoin, None],
                        "target_vector",
                        target_vector,
                    ),
                ]
            )

        if isinstance(near_text, str):
            near_text = [near_text]
        certainty, distance = self._parse_near_options(certainty, distance)
        targets, target_vector = self.__target_vector_to_grpc(target_vector)

        return base_search_pb2.NearTextSearch(
            query=near_text,
            certainty=certainty,
            distance=distance,
            move_away=self.__parse_move(move_away),
            move_to=self.__parse_move(move_to),
            targets=targets,
            target_vectors=target_vector,
        )

    def _parse_near_object(
        self,
        near_object: UUID,
        certainty: Optional[NUMBER],
        distance: Optional[NUMBER],
        target_vector: Optional[TargetVectorJoinType],
    ) -> base_search_pb2.NearObject:
        if self._validate_arguments:
            _validate_input(
                [
                    _ValidateArgument([str, uuid_lib.UUID], "near_object", near_object),
                    _ValidateArgument(
                        [str, None, List, _MultiTargetVectorJoin],
                        "target_vector",
                        target_vector,
                    ),
                ]
            )

        certainty, distance = self._parse_near_options(certainty, distance)

        targets, target_vector = self.__target_vector_to_grpc(target_vector)

        return base_search_pb2.NearObject(
            id=str(near_object),
            certainty=certainty,
            distance=distance,
            targets=targets,
            target_vectors=target_vector,
        )

    def _parse_media(
        self,
        media: str,
        type_: Literal["audio", "depth", "image", "imu", "thermal", "video"],
        certainty: Optional[NUMBER],
        distance: Optional[NUMBER],
        target_vector: Optional[TargetVectorJoinType],
    ) -> dict:
        if self._validate_arguments:
            _validate_input(
                [
                    _ValidateArgument([str], "media", media),
                    _ValidateArgument(
                        [str, None, List, _MultiTargetVectorJoin],
                        "target_vector",
                        target_vector,
                    ),
                ]
            )

        certainty, distance = self._parse_near_options(certainty, distance)

        kwargs: Dict[str, Any] = {}
        targets, target_vector = self.__target_vector_to_grpc(target_vector)
        if type_ == "audio":
            kwargs["near_audio"] = base_search_pb2.NearAudioSearch(
                audio=media,
                distance=distance,
                certainty=certainty,
                target_vectors=target_vector,
                targets=targets,
            )
        elif type_ == "depth":
            kwargs["near_depth"] = base_search_pb2.NearDepthSearch(
                depth=media,
                distance=distance,
                certainty=certainty,
                target_vectors=target_vector,
                targets=targets,
            )
        elif type_ == "image":
            kwargs["near_image"] = base_search_pb2.NearImageSearch(
                image=media,
                distance=distance,
                certainty=certainty,
                target_vectors=target_vector,
                targets=targets,
            )
        elif type_ == "imu":
            kwargs["near_imu"] = base_search_pb2.NearIMUSearch(
                imu=media,
                distance=distance,
                certainty=certainty,
                target_vectors=target_vector,
                targets=targets,
            )
        elif type_ == "thermal":
            kwargs["near_thermal"] = base_search_pb2.NearThermalSearch(
                thermal=media,
                distance=distance,
                certainty=certainty,
                target_vectors=target_vector,
                targets=targets,
            )
        elif type_ == "video":
            kwargs["near_video"] = base_search_pb2.NearVideoSearch(
                video=media,
                distance=distance,
                certainty=certainty,
                target_vectors=target_vector,
                targets=targets,
            )
        else:
            raise ValueError(
                f"type_ must be one of ['audio', 'depth', 'image', 'imu', 'thermal', 'video'], but got {type_}"
            )
        return kwargs

    def _parse_hybrid(
        self,
        query: Optional[str],
        alpha: Optional[float],
        vector: Optional[HybridVectorType],
        properties: Optional[List[str]],
        bm25_operator: Optional[BM25OperatorOptions],
        fusion_type: Optional[HybridFusion],
        distance: Optional[NUMBER],
        target_vector: Optional[TargetVectorJoinType],
    ) -> Union[base_search_pb2.Hybrid, None]:
        if self._validate_arguments:
            _validate_input(
                [
                    _ValidateArgument([None, str], "query", query),
                    _ValidateArgument([float, int, None], "alpha", alpha),
                    _ValidateArgument(
                        [
                            List,
                            Dict,
                            _ExtraTypes.PANDAS,
                            _ExtraTypes.POLARS,
                            _ExtraTypes.NUMPY,
                            _ExtraTypes.TF,
                            _HybridNearText,
                            _HybridNearVector,
                            None,
                        ],
                        "vector",
                        vector,
                    ),
                    _ValidateArgument([List, None], "properties", properties),
                    _ValidateArgument([HybridFusion, None], "fusion_type", fusion_type),
                    _ValidateArgument(
                        [str, None, List, _MultiTargetVectorJoin],
                        "target_vector",
                        target_vector,
                    ),
                ]
            )

        # Set hybrid search to only query the other search-type if one of the two is not set
        if query is None:
            alpha = 1

        targets, target_vectors = self.__target_vector_to_grpc(target_vector)

        near_text, near_vector, vector_bytes, vectors = None, None, None, None

        if vector is None:
            pass
        elif isinstance(vector, list) and len(vector) > 0 and isinstance(vector[0], float):
            # fast path for simple vector
            vector_bytes = struct.pack("{}f".format(len(vector)), *vector)
        elif _is_2d_vector(vector) and self._weaviate_version.is_at_least(1, 29, 0):
            # fast path for simple multi-vector
            vectors = [
                base_pb2.Vectors(
                    vector_bytes=_Pack.multi(vector),
                    type=base_pb2.Vectors.VECTOR_TYPE_MULTI_FP32,
                )
            ]
        elif isinstance(vector, _HybridNearText):
            near_text = base_search_pb2.NearTextSearch(
                query=[vector.text] if isinstance(vector.text, str) else vector.text,
                certainty=vector.certainty,
                distance=vector.distance,
                move_away=self.__parse_move(vector.move_away),
                move_to=self.__parse_move(vector.move_to),
            )
        elif isinstance(vector, _HybridNearVector):
            if self._weaviate_version.is_lower_than(1, 27, 0):
                vector_per_target_tmp, vector_bytes_tmp = self._vector_per_target(
                    vector.vector, targets, "vector"
                )
                vector_for_targets_tmp = None
            else:
                (
                    vector_for_targets_tmp,
                    vector_bytes_tmp,
                    target_vectors_tmp,
                ) = self._vector_for_target(vector.vector, targets, "vector")
                vector_per_target_tmp = None
                if target_vectors_tmp is not None:
                    targets, target_vectors = self._recompute_target_vector_to_grpc(
                        target_vector, target_vectors_tmp
                    )

            near_vector = base_search_pb2.NearVector(
                vector_bytes=vector_bytes_tmp,
                certainty=vector.certainty,
                distance=vector.distance,
                vector_per_target=vector_per_target_tmp,
                vector_for_targets=vector_for_targets_tmp,
            )
        else:
            if self._weaviate_version.is_lower_than(1, 27, 0):
                vector_per_target_tmp, vector_bytes_tmp = self._vector_per_target(
                    vector, targets, "vector"
                )
                vector_for_targets_tmp = None
            else:
                (
                    vector_for_targets_tmp,
                    vector_bytes_tmp,
                    target_vectors_tmp,
                ) = self._vector_for_target(vector, targets, "vector")
                vector_per_target_tmp = None
                if target_vectors_tmp is not None:
                    targets, target_vectors = self._recompute_target_vector_to_grpc(
                        target_vector, target_vectors_tmp
                    )
                else:
                    targets, target_vectors = self.__target_vector_to_grpc(target_vector)

            if vector_per_target_tmp is not None or vector_for_targets_tmp is not None:
                near_vector = base_search_pb2.NearVector(
                    vector_bytes=vector_bytes_tmp,
                    vector_per_target=vector_per_target_tmp,
                    vector_for_targets=vector_for_targets_tmp,
                )
            else:
                vector_bytes = vector_bytes_tmp

        return (
            base_search_pb2.Hybrid(
                properties=properties,
                query=query,
                alpha=float(alpha) if alpha is not None else None,
                fusion_type=(
                    cast(
                        base_search_pb2.Hybrid.FusionType,
                        base_search_pb2.Hybrid.FusionType.Value(fusion_type.value),
                    )
                    if fusion_type is not None
                    else None
                ),
                target_vectors=target_vectors,
                targets=targets,
                near_text=near_text,
                near_vector=near_vector,
                vector_bytes=vector_bytes,
                vector_distance=distance,
                vectors=vectors,
                bm25_search_operator=base_search_pb2.SearchOperatorOptions(
                    operator=bm25_operator.operator,
                    minimum_or_tokens_match=bm25_operator.minimum_should_match
                    if isinstance(bm25_operator, BM25OperatorOr)
                    else None,
                )
                if bm25_operator is not None
                else None,
            )
            if query is not None or vector is not None
            else None
        )


class _ByteOps:
    @staticmethod
    def decode_float32s(byte_vector: bytes) -> List[float]:
        return [
            float(val) for val in struct.unpack(f"{len(byte_vector) // UINT32_LEN}f", byte_vector)
        ]

    @staticmethod
    def decode_float64s(byte_vector: bytes) -> List[float]:
        return [
            float(val) for val in struct.unpack(f"{len(byte_vector) // UINT64_LEN}d", byte_vector)
        ]

    @staticmethod
    def decode_int64s(byte_vector: bytes) -> List[int]:
        return [
            int(val) for val in struct.unpack(f"{len(byte_vector) // UINT64_LEN}q", byte_vector)
        ]


@dataclass
class _Packing:
    bytes_: bytes
    type_: base_pb2.Vectors.VectorType


class _Pack:
    @staticmethod
    def parse_single_or_multi_vec(vector: PrimitiveVectorType) -> _Packing:
        if _is_2d_vector(vector):
            return _Packing(
                bytes_=_Pack.multi(vector),
                type_=base_pb2.Vectors.VECTOR_TYPE_MULTI_FP32,
            )
        elif _is_1d_vector(vector):
            return _Packing(
                bytes_=_Pack.single(vector),
                type_=base_pb2.Vectors.VECTOR_TYPE_SINGLE_FP32,
            )
        else:
            raise WeaviateInvalidInputError(f"Invalid vectors: {vector}")

    @staticmethod
    def single(vector: OneDimensionalVectorType) -> bytes:
        vector_list = _get_vector_v4(vector)
        return struct.pack("{}f".format(len(vector_list)), *vector_list)

    @staticmethod
    def multi(vector: TwoDimensionalVectorType) -> bytes:
        vector_list = [item for sublist in vector for item in sublist]
        return struct.pack("<H", len(vector[0])) + struct.pack(
            "{}f".format(len(vector_list)), *vector_list
        )


class _Unpack:
    @staticmethod
    def single(byte_vector: bytes) -> List[float]:
        return _ByteOps.decode_float32s(byte_vector)

    @staticmethod
    def multi(byte_vector: bytes) -> List[List[float]]:
        dim_bytes = byte_vector[:2]
        dim = int(struct.unpack("<H", dim_bytes)[0])
        byte_vector = byte_vector[2:]
        how_many = len(byte_vector) // (dim * UINT32_LEN)
        return [
            _ByteOps.decode_float32s(byte_vector[i * dim * UINT32_LEN : (i + 1) * dim * UINT32_LEN])
            for i in range(how_many)
        ]


def _is_1d_vector(inputs: Any) -> TypeGuard[OneDimensionalVectorType]:
    try:
        if len(inputs) == 0:
            return False
    except TypeError:
        return False
    if __is_list_type(inputs):
        return not __is_list_type(inputs[0])
    return False


def _is_2d_vector(inputs: Any) -> TypeGuard[TwoDimensionalVectorType]:
    try:
        if len(inputs) == 0:
            return False
    except TypeError:
        return False
    if __is_list_type(inputs):
        return _is_1d_vector(inputs[0])
    return False


def __is_list_type(inputs: Any) -> bool:
    try:
        if len(inputs) == 0:
            return False
    except TypeError:
        return False

    return any(
        _is_valid(types, inputs)
        for types in [
            List,
            _ExtraTypes.TF,
            _ExtraTypes.PANDAS,
            _ExtraTypes.NUMPY,
            _ExtraTypes.POLARS,
        ]
    )



import uuid
from typing import List, Union

import pytest

from integration.conftest import ClientFactory, CollectionFactory
from weaviate.collections.classes.config import (
    Configure,
    DataType,
    Property,
    Reconfigure,
)
from weaviate.collections.classes.data import (
    DataObject,
)
from weaviate.collections.classes.tenants import (
    Tenant,
    TenantActivityStatus,
    TenantCreate,
)
from weaviate.collections.tenants import TenantCreateInputType
from weaviate.exceptions import (
    WeaviateInvalidInputError,
    WeaviateUnsupportedFeatureError,
)


def test_shards_on_tenants(
    client_factory: ClientFactory, collection_factory: CollectionFactory
) -> None:
    collection = collection_factory(
        vectorizer_config=Configure.Vectorizer.none(),
        multi_tenancy_config=Configure.multi_tenancy(enabled=True),
    )
    collection.tenants.create(Tenant(name="tenant1"))
    client = client_factory()

    nodes = client.cluster.nodes(collection.name, output="verbose")
    count = sum(len(node.shards) for node in nodes)

    assert count == 1


@pytest.mark.parametrize("tenant", ["tenant1", Tenant(name="tenant1")])
def test_delete_by_id_tenant(
    collection_factory: CollectionFactory, tenant: Union[str, Tenant]
) -> None:
    collection = collection_factory(
        vectorizer_config=Configure.Vectorizer.none(),
        multi_tenancy_config=Configure.multi_tenancy(enabled=True),
    )
    collection.tenants.create(tenant)
    tenant1 = collection.with_tenant(tenant)
    uuid = tenant1.data.insert(properties={})
    assert tenant1.query.fetch_object_by_id(uuid) is not None
    assert tenant1.data.delete_by_id(uuid)
    assert tenant1.query.fetch_object_by_id(uuid) is None


def test_insert_many_with_tenant(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        properties=[Property(name="Name", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.none(),
        multi_tenancy_config=Configure.multi_tenancy(enabled=True),
    )

    collection.tenants.create([Tenant(name="tenant1"), Tenant(name="tenant2")])
    tenant1 = collection.with_tenant("tenant1")
    tenant2 = collection.with_tenant("tenant2")

    ret = tenant1.data.insert_many(
        [
            DataObject(properties={"name": "some name"}, vector=[1, 2, 3]),
            DataObject(properties={"name": "some other name"}, uuid=uuid.uuid4()),
        ]
    )
    assert not ret.has_errors
    obj1 = tenant1.query.fetch_object_by_id(ret.uuids[0])
    obj2 = tenant1.query.fetch_object_by_id(ret.uuids[1])
    assert obj1.properties["name"] == "some name"
    assert obj2.properties["name"] == "some other name"
    assert tenant2.query.fetch_object_by_id(ret.uuids[0]) is None
    assert tenant2.query.fetch_object_by_id(ret.uuids[1]) is None


def test_replace_with_tenant(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        properties=[Property(name="Name", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.none(),
        multi_tenancy_config=Configure.multi_tenancy(enabled=True),
    )

    collection.tenants.create([Tenant(name="tenant1"), Tenant(name="tenant2")])
    tenant1 = collection.with_tenant("tenant1")
    tenant2 = collection.with_tenant("tenant2")

    uuid = tenant1.data.insert(properties={"name": "some name"})
    tenant1.data.replace(properties={"name": "other name"}, uuid=uuid)
    assert tenant1.query.fetch_object_by_id(uuid).properties["name"] == "other name"
    assert tenant2.query.fetch_object_by_id(uuid) is None


def test_tenants_update(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        properties=[Property(name="Name", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    uuid = collection.data.insert(properties={"name": "some name"})
    collection.data.update(properties={"name": "other name"}, uuid=uuid, vector=[1, 2, 3])
    obj = collection.query.fetch_object_by_id(uuid, include_vector=True)
    assert obj.properties["name"] == "other name"
    assert obj.vector["default"] == [1, 2, 3]


def test_update_with_tenant(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        properties=[Property(name="Name", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.none(),
        multi_tenancy_config=Configure.multi_tenancy(enabled=True),
    )

    collection.tenants.create([Tenant(name="tenant1"), Tenant(name="tenant2")])
    tenant1 = collection.with_tenant("tenant1")
    tenant2 = collection.with_tenant("tenant2")

    uuid = tenant1.data.insert(properties={"name": "some name"})
    tenant1.data.update(properties={"name": "other name"}, uuid=uuid)
    assert tenant1.query.fetch_object_by_id(uuid).properties["name"] == "other name"
    assert tenant2.query.fetch_object_by_id(uuid) is None


def test_tenants(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        vectorizer_config=Configure.Vectorizer.none(),
        multi_tenancy_config=Configure.multi_tenancy(),
    )

    collection.tenants.create([Tenant(name="tenant1"), Tenant(name="tenant2")])

    tenants = collection.tenants.get()
    assert len(tenants) == 2
    assert isinstance(tenants["tenant1"], Tenant)
    assert isinstance(tenants["tenant2"], Tenant)
    assert tenants["tenant1"].name == "tenant1"
    assert tenants["tenant2"].name == "tenant2"

    if collection._connection._weaviate_version.supports_tenants_get_grpc:
        tenants = collection.tenants.get_by_names(tenants=["tenant2"])
        assert len(tenants) == 1
        assert isinstance(tenants["tenant2"], Tenant)
        assert tenants["tenant2"].name == "tenant2"
    else:
        pytest.raises(
            WeaviateUnsupportedFeatureError, collection.tenants.get_by_names, tenants=["tenant2"]
        )

    collection.tenants.remove(["tenant1", "tenant2"])

    tenants = collection.tenants.get()
    assert len(tenants) == 0


def test_search_with_tenant(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[Property(name="name", data_type=DataType.TEXT)],
        multi_tenancy_config=Configure.multi_tenancy(enabled=True),
    )

    collection.tenants.create([Tenant(name="Tenant1"), Tenant(name="Tenant2")])
    tenant1 = collection.with_tenant("Tenant1")
    tenant2 = collection.with_tenant("Tenant2")
    uuid1 = tenant1.data.insert({"name": "some name"})
    objects1 = tenant1.query.bm25(query="some").objects
    assert len(objects1) == 1
    assert objects1[0].uuid == uuid1

    objects2 = tenant2.query.bm25(query="some").objects
    assert len(objects2) == 0


def test_fetch_object_by_id_with_tenant(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[Property(name="name", data_type=DataType.TEXT)],
        multi_tenancy_config=Configure.multi_tenancy(enabled=True),
    )

    collection.tenants.create([Tenant(name="Tenant1"), Tenant(name="Tenant2")])
    tenant1 = collection.with_tenant("Tenant1")
    tenant2 = collection.with_tenant("Tenant2")

    uuid1 = tenant1.data.insert({"name": "some name"})
    obj1 = tenant1.query.fetch_object_by_id(uuid1)
    assert obj1.properties["name"] == "some name"

    obj2 = tenant2.query.fetch_object_by_id(uuid1)
    assert obj2 is None

    uuid2 = tenant2.data.insert({"name": "some other name"})
    obj3 = tenant2.query.fetch_object_by_id(uuid2)
    assert obj3.properties["name"] == "some other name"

    obj4 = tenant1.query.fetch_object_by_id(uuid2)
    assert obj4 is None


def test_fetch_objects_with_tenant(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[Property(name="name", data_type=DataType.TEXT)],
        multi_tenancy_config=Configure.multi_tenancy(enabled=True),
    )

    collection.tenants.create([Tenant(name="Tenant1"), Tenant(name="Tenant2")])
    tenant1 = collection.with_tenant("Tenant1")
    tenant2 = collection.with_tenant("Tenant2")

    tenant1.data.insert({"name": "some name"})
    objects = tenant1.query.fetch_objects().objects
    assert len(objects) == 1
    assert objects[0].properties["name"] == "some name"

    objects = tenant2.query.fetch_objects().objects
    assert len(objects) == 0

    tenant2.data.insert({"name": "some other name"})
    objects = tenant2.query.fetch_objects().objects
    assert len(objects) == 1
    assert objects[0].properties["name"] == "some other name"


def test_exist_with_tenant(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        vectorizer_config=Configure.Vectorizer.none(),
        multi_tenancy_config=Configure.multi_tenancy(enabled=True),
    )
    collection.tenants.create([Tenant(name="Tenant1"), Tenant(name="Tenant2")])

    uuid1 = collection.with_tenant("Tenant1").data.insert({})
    uuid2 = collection.with_tenant("Tenant2").data.insert({})

    assert collection.with_tenant("Tenant1").data.exists(uuid1)
    assert not collection.with_tenant("Tenant2").data.exists(uuid1)
    assert collection.with_tenant("Tenant2").data.exists(uuid2)
    assert not collection.with_tenant("Tenant1").data.exists(uuid2)


def test_tenant_with_activity(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        vectorizer_config=Configure.Vectorizer.none(),
        multi_tenancy_config=Configure.multi_tenancy(enabled=True),
    )

    with pytest.warns(DeprecationWarning) as recwarn:
        collection.tenants.create(
            [
                Tenant(name="1", activity_status=TenantActivityStatus.HOT),
                Tenant(name="2", activity_status=TenantActivityStatus.COLD),
                Tenant(name="3", activity_status=TenantActivityStatus.ACTIVE),
                Tenant(name="4", activity_status=TenantActivityStatus.INACTIVE),
                Tenant(name="5"),
            ]
        )
        assert len(recwarn) == 2
        assert any("HOT is deprecated" in warn.message.args[0] for warn in recwarn.list)
        assert any("COLD is deprecated" in warn.message.args[0] for warn in recwarn.list)
    tenants = collection.tenants.get()
    assert tenants["1"].activity_status == TenantActivityStatus.ACTIVE
    assert tenants["2"].activity_status == TenantActivityStatus.INACTIVE
    assert tenants["3"].activity_status == TenantActivityStatus.ACTIVE
    assert tenants["4"].activity_status == TenantActivityStatus.INACTIVE
    assert tenants["5"].activity_status == TenantActivityStatus.ACTIVE


def test_update_tenant(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        vectorizer_config=Configure.Vectorizer.none(),
        multi_tenancy_config=Configure.multi_tenancy(enabled=True),
    )
    with pytest.warns(DeprecationWarning) as recwarn:
        collection.tenants.create(Tenant(name="1", activity_status=TenantActivityStatus.HOT))
        assert len(recwarn) == 1
        assert any("HOT is deprecated" in warn.message.args[0] for warn in recwarn.list)
    tenants = collection.tenants.get()
    assert tenants["1"].activity_status == TenantActivityStatus.ACTIVE

    with pytest.warns(DeprecationWarning) as recwarn:
        collection.tenants.update(Tenant(name="1", activity_status=TenantActivityStatus.COLD))
        assert len(recwarn) == 1
        assert any("COLD is deprecated" in warn.message.args[0] for warn in recwarn.list)
    tenants = collection.tenants.get()
    assert tenants["1"].activity_status == TenantActivityStatus.INACTIVE

    collection.tenants.update(Tenant(name="1", activity_status=TenantActivityStatus.ACTIVE))
    tenants = collection.tenants.get()
    assert tenants["1"].activity_status == TenantActivityStatus.ACTIVE

    collection.tenants.update(Tenant(name="1", activity_status=TenantActivityStatus.INACTIVE))
    tenants = collection.tenants.get()
    assert tenants["1"].activity_status == TenantActivityStatus.INACTIVE


def test_tenant_exists(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        vectorizer_config=Configure.Vectorizer.none(),
        multi_tenancy_config=Configure.multi_tenancy(enabled=True),
    )
    tenant = Tenant(name="1")
    collection.tenants.create([tenant])

    if collection._connection._weaviate_version.is_lower_than(1, 25, 0):
        with pytest.raises(WeaviateUnsupportedFeatureError):
            collection.tenants.exists(tenant.name)
    else:
        assert collection.tenants.exists(tenant.name)
        assert collection.tenants.exists(tenant)
        assert not collection.tenants.exists("2")


@pytest.mark.parametrize("tenant1", ["tenant1", Tenant(name="tenant1")])
@pytest.mark.parametrize("tenant2", ["tenant2", Tenant(name="tenant2")])
@pytest.mark.parametrize("tenant3", ["tenant3", Tenant(name="tenant3")])
def test_tenant_get_by_name(
    collection_factory: CollectionFactory,
    tenant1: Union[str, Tenant],
    tenant2: Union[str, Tenant],
    tenant3: Union[str, Tenant],
) -> None:
    collection = collection_factory(
        vectorizer_config=Configure.Vectorizer.none(),
        multi_tenancy_config=Configure.multi_tenancy(),
    )

    collection.tenants.create([Tenant(name="tenant1")])

    if collection._connection._weaviate_version.supports_tenants_get_grpc:
        tenant = collection.tenants.get_by_name(tenant1)
        assert tenant is not None
        assert tenant.name == "tenant1"

        tenant = collection.tenants.get_by_name(tenant2)
        assert tenant is None
    else:
        pytest.raises(WeaviateUnsupportedFeatureError, collection.tenants.get_by_name, tenant3)


def test_autotenant_toggling(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        vectorizer_config=Configure.Vectorizer.none(),
        multi_tenancy_config=Configure.multi_tenancy(enabled=True),
    )
    if collection._connection._weaviate_version.is_lower_than(1, 25, 0):
        pytest.skip("Auto-tenant creation is not supported in this version")

    assert not collection.config.get().multi_tenancy_config.auto_tenant_creation

    collection.config.update(
        multi_tenancy_config=Reconfigure.multi_tenancy(auto_tenant_creation=True)
    )
    assert collection.config.get().multi_tenancy_config.auto_tenant_creation

    collection.config.update(
        multi_tenancy_config=Reconfigure.multi_tenancy(auto_tenant_creation=False)
    )
    assert not collection.config.get().multi_tenancy_config.auto_tenant_creation


@pytest.mark.parametrize(
    "tenants",
    [
        "tenant",
        Tenant(name="tenant"),
        TenantCreate(name="tenant"),
        ["tenant"],
        [Tenant(name="tenant")],
        [TenantCreate(name="tenant")],
    ],
)
def test_tenants_create(
    collection_factory: CollectionFactory,
    tenants: Union[TenantCreateInputType, List[TenantCreateInputType]],
) -> None:
    collection = collection_factory(
        vectorizer_config=Configure.Vectorizer.none(),
        multi_tenancy_config=Configure.multi_tenancy(),
    )

    collection.tenants.create(tenants)
    t = collection.tenants.get()
    assert len(t) == 1
    assert t["tenant"].name == "tenant"


@pytest.mark.parametrize(
    "tenants",
    [
        "tenant",
        Tenant(name="tenant"),
        ["tenant"],
        [Tenant(name="tenant")],
    ],
)
def test_tenants_remove(
    collection_factory: CollectionFactory, tenants: Union[str, Tenant, List[Union[str, Tenant]]]
) -> None:
    collection = collection_factory(
        vectorizer_config=Configure.Vectorizer.none(),
        multi_tenancy_config=Configure.multi_tenancy(),
    )

    collection.tenants.create(["tenant", "tenantt"])
    collection.tenants.remove(tenants)
    t = collection.tenants.get()
    assert "tenant" not in t
    assert "tenantt" in t


@pytest.mark.parametrize(
    "tenants",
    [
        Tenant(name="1", activity_status=TenantActivityStatus.FROZEN),
        [
            Tenant(name="4", activity_status=TenantActivityStatus.FROZEN),
        ],
    ],
)
def test_tenants_create_with_read_only_activity_status(
    collection_factory: CollectionFactory, tenants: Union[Tenant, List[Tenant]]
) -> None:
    collection = collection_factory(
        vectorizer_config=Configure.Vectorizer.none(),
        multi_tenancy_config=Configure.multi_tenancy(),
    )
    with pytest.raises(WeaviateInvalidInputError):
        collection.tenants.create(tenants)


@pytest.mark.parametrize(
    "tenants",
    [
        Tenant(name="1", activity_status=TenantActivityStatus.OFFLOADING),
        Tenant(name="1", activity_status=TenantActivityStatus.ONLOADING),
        [
            Tenant(name="1", activity_status=TenantActivityStatus.OFFLOADING),
            Tenant(name="2", activity_status=TenantActivityStatus.ONLOADING),
        ],
    ],
)
def test_tenants_update_with_read_only_activity_status(
    collection_factory: CollectionFactory, tenants: Union[Tenant, List[Tenant]]
) -> None:
    collection = collection_factory(
        vectorizer_config=Configure.Vectorizer.none(),
        multi_tenancy_config=Configure.multi_tenancy(),
    )
    with pytest.raises(WeaviateInvalidInputError):
        collection.tenants.update(tenants)


def test_tenants_create_and_update_1001_tenants(
    collection_factory: CollectionFactory,
) -> None:
    collection = collection_factory(
        vectorizer_config=Configure.Vectorizer.none(),
        multi_tenancy_config=Configure.multi_tenancy(),
    )

    tenants = [TenantCreate(name=f"tenant{i}") for i in range(1001)]

    collection.tenants.create(tenants)
    t = collection.tenants.get()
    assert len(t) == 1001
    assert all(tenant.activity_status == TenantActivityStatus.ACTIVE for tenant in t.values())

    tenants = [
        Tenant(name=f"tenant{i}", activity_status=TenantActivityStatus.INACTIVE)
        for i in range(1001)
    ]
    collection.tenants.update(tenants)
    t = collection.tenants.get()
    assert len(t) == 1001
    assert all(tenant.activity_status == TenantActivityStatus.INACTIVE for tenant in t.values())


def test_tenants_auto_tenant_creation(
    client_factory: ClientFactory, collection_factory: CollectionFactory
) -> None:
    dummy = collection_factory("dummy")
    if dummy._connection._weaviate_version.is_lower_than(1, 25, 0):
        pytest.skip("Auto-tenant creation is not supported in this version")

    collection = collection_factory(
        properties=[Property(name="name", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.none(),
        multi_tenancy_config=Configure.multi_tenancy(auto_tenant_creation=True),
    )

    collection.with_tenant("tenant").data.insert_many(
        [DataObject(properties={"name": "some name"}) for _ in range(101)]
    )

    client = client_factory()
    with client.batch.fixed_size(batch_size=101) as batch:
        for i in range(101):
            batch.add_object(
                collection=collection.name, properties={"name": "some name"}, tenant=f"tenant-{i}"
            )
    assert len(client.batch.failed_objects) == 0


def test_tenants_deactivate_then_activate(collection_factory: CollectionFactory) -> None:
    collection = collection_factory(
        properties=[Property(name="name", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.none(),
        multi_tenancy_config=Configure.multi_tenancy(enabled=True),
    )

    t = Tenant(name="tenant1")
    collection.tenants.create([t])
    tenant1 = collection.with_tenant(t)

    tenant1.data.insert(properties={"name": "some name"})
    assert tenant1.query.fetch_objects().objects[0].properties["name"] == "some name"

    collection.tenants.deactivate(t)
    assert collection.tenants.get()["tenant1"].activity_status == TenantActivityStatus.INACTIVE

    collection.tenants.activate(t)
    assert tenant1.query.fetch_objects().objects[0].properties["name"] == "some name"



from typing import Optional, Sequence

from weaviate.collections.classes.tenants import TenantActivityStatus
from weaviate.collections.grpc.shared import _BaseGRPC
from weaviate.proto.v1 import tenants_pb2
from weaviate.util import _ServerVersion


class _TenantsGRPC(_BaseGRPC):
    def __init__(
        self,
        weaviate_version: _ServerVersion,
        name: str,
    ):
        super().__init__(weaviate_version, None, False)
        self._name: str = name

    def get(self, names: Optional[Sequence[str]]) -> tenants_pb2.TenantsGetRequest:
        return tenants_pb2.TenantsGetRequest(
            collection=self._name,
            names=tenants_pb2.TenantNames(values=names) if names is not None else None,
        )

    def map_activity_status(self, status: tenants_pb2.TenantActivityStatus) -> TenantActivityStatus:
        if (
            status == tenants_pb2.TENANT_ACTIVITY_STATUS_COLD
            or status == tenants_pb2.TENANT_ACTIVITY_STATUS_INACTIVE
        ):
            return TenantActivityStatus.INACTIVE
        if (
            status == tenants_pb2.TENANT_ACTIVITY_STATUS_HOT
            or status == tenants_pb2.TENANT_ACTIVITY_STATUS_ACTIVE
        ):
            return TenantActivityStatus.ACTIVE
        if (
            status == tenants_pb2.TENANT_ACTIVITY_STATUS_FROZEN
            or status == tenants_pb2.TENANT_ACTIVITY_STATUS_OFFLOADED
        ):
            return TenantActivityStatus.OFFLOADED
        if (
            status == tenants_pb2.TENANT_ACTIVITY_STATUS_FREEZING
            or status == tenants_pb2.TENANT_ACTIVITY_STATUS_OFFLOADING
        ):
            return TenantActivityStatus.OFFLOADING
        if (
            status == tenants_pb2.TENANT_ACTIVITY_STATUS_UNFREEZING
            or status == tenants_pb2.TENANT_ACTIVITY_STATUS_ONLOADING
        ):
            return TenantActivityStatus.ONLOADING
        raise ValueError(f"Unknown TenantActivityStatus: {status}")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: v1/weaviate.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from weaviate.proto.v1.v4216.v1 import aggregate_pb2 as v1_dot_aggregate__pb2
from weaviate.proto.v1.v4216.v1 import batch_pb2 as v1_dot_batch__pb2
from weaviate.proto.v1.v4216.v1 import batch_delete_pb2 as v1_dot_batch__delete__pb2
from weaviate.proto.v1.v4216.v1 import search_get_pb2 as v1_dot_search__get__pb2
from weaviate.proto.v1.v4216.v1 import tenants_pb2 as v1_dot_tenants__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11v1/weaviate.proto\x12\x0bweaviate.v1\x1a\x12v1/aggregate.proto\x1a\x0ev1/batch.proto\x1a\x15v1/batch_delete.proto\x1a\x13v1/search_get.proto\x1a\x10v1/tenants.proto2\x87\x05\n\x08Weaviate\x12@\n\x06Search\x12\x1a.weaviate.v1.SearchRequest\x1a\x18.weaviate.v1.SearchReply\"\x00\x12R\n\x0c\x42\x61tchObjects\x12 .weaviate.v1.BatchObjectsRequest\x1a\x1e.weaviate.v1.BatchObjectsReply\"\x00\x12[\n\x0f\x42\x61tchReferences\x12#.weaviate.v1.BatchReferencesRequest\x1a!.weaviate.v1.BatchReferencesReply\"\x00\x12O\n\x0b\x42\x61tchDelete\x12\x1f.weaviate.v1.BatchDeleteRequest\x1a\x1d.weaviate.v1.BatchDeleteReply\"\x00\x12L\n\nTenantsGet\x12\x1e.weaviate.v1.TenantsGetRequest\x1a\x1c.weaviate.v1.TenantsGetReply\"\x00\x12I\n\tAggregate\x12\x1d.weaviate.v1.AggregateRequest\x1a\x1b.weaviate.v1.AggregateReply\"\x00\x12I\n\tBatchSend\x12\x1d.weaviate.v1.BatchSendRequest\x1a\x1b.weaviate.v1.BatchSendReply\"\x00\x12S\n\x0b\x42\x61tchStream\x12\x1f.weaviate.v1.BatchStreamRequest\x1a\x1f.weaviate.v1.BatchStreamMessage\"\x00\x30\x01\x42j\n#io.weaviate.client.grpc.protocol.v1B\rWeaviateProtoZ4github.com/weaviate/weaviate/grpc/generated;protocolb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'v1.weaviate_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n#io.weaviate.client.grpc.protocol.v1B\rWeaviateProtoZ4github.com/weaviate/weaviate/grpc/generated;protocol'
  _globals['_WEAVIATE']._serialized_start=133
  _globals['_WEAVIATE']._serialized_end=780
# @@protoc_insertion_point(module_scope)



from google.protobuf import struct_pb2 as _struct_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Properties(_message.Message):
    __slots__ = ("fields",)
    class FieldsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: Value
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[Value, _Mapping]] = ...) -> None: ...
    FIELDS_FIELD_NUMBER: _ClassVar[int]
    fields: _containers.MessageMap[str, Value]
    def __init__(self, fields: _Optional[_Mapping[str, Value]] = ...) -> None: ...

class Value(_message.Message):
    __slots__ = ("number_value", "bool_value", "object_value", "list_value", "date_value", "uuid_value", "int_value", "geo_value", "blob_value", "phone_value", "null_value", "text_value")
    NUMBER_VALUE_FIELD_NUMBER: _ClassVar[int]
    BOOL_VALUE_FIELD_NUMBER: _ClassVar[int]
    OBJECT_VALUE_FIELD_NUMBER: _ClassVar[int]
    LIST_VALUE_FIELD_NUMBER: _ClassVar[int]
    DATE_VALUE_FIELD_NUMBER: _ClassVar[int]
    UUID_VALUE_FIELD_NUMBER: _ClassVar[int]
    INT_VALUE_FIELD_NUMBER: _ClassVar[int]
    GEO_VALUE_FIELD_NUMBER: _ClassVar[int]
    BLOB_VALUE_FIELD_NUMBER: _ClassVar[int]
    PHONE_VALUE_FIELD_NUMBER: _ClassVar[int]
    NULL_VALUE_FIELD_NUMBER: _ClassVar[int]
    TEXT_VALUE_FIELD_NUMBER: _ClassVar[int]
    number_value: float
    bool_value: bool
    object_value: Properties
    list_value: ListValue
    date_value: str
    uuid_value: str
    int_value: int
    geo_value: GeoCoordinate
    blob_value: str
    phone_value: PhoneNumber
    null_value: _struct_pb2.NullValue
    text_value: str
    def __init__(self, number_value: _Optional[float] = ..., bool_value: bool = ..., object_value: _Optional[_Union[Properties, _Mapping]] = ..., list_value: _Optional[_Union[ListValue, _Mapping]] = ..., date_value: _Optional[str] = ..., uuid_value: _Optional[str] = ..., int_value: _Optional[int] = ..., geo_value: _Optional[_Union[GeoCoordinate, _Mapping]] = ..., blob_value: _Optional[str] = ..., phone_value: _Optional[_Union[PhoneNumber, _Mapping]] = ..., null_value: _Optional[_Union[_struct_pb2.NullValue, str]] = ..., text_value: _Optional[str] = ...) -> None: ...

class ListValue(_message.Message):
    __slots__ = ("number_values", "bool_values", "object_values", "date_values", "uuid_values", "int_values", "text_values")
    NUMBER_VALUES_FIELD_NUMBER: _ClassVar[int]
    BOOL_VALUES_FIELD_NUMBER: _ClassVar[int]
    OBJECT_VALUES_FIELD_NUMBER: _ClassVar[int]
    DATE_VALUES_FIELD_NUMBER: _ClassVar[int]
    UUID_VALUES_FIELD_NUMBER: _ClassVar[int]
    INT_VALUES_FIELD_NUMBER: _ClassVar[int]
    TEXT_VALUES_FIELD_NUMBER: _ClassVar[int]
    number_values: NumberValues
    bool_values: BoolValues
    object_values: ObjectValues
    date_values: DateValues
    uuid_values: UuidValues
    int_values: IntValues
    text_values: TextValues
    def __init__(self, number_values: _Optional[_Union[NumberValues, _Mapping]] = ..., bool_values: _Optional[_Union[BoolValues, _Mapping]] = ..., object_values: _Optional[_Union[ObjectValues, _Mapping]] = ..., date_values: _Optional[_Union[DateValues, _Mapping]] = ..., uuid_values: _Optional[_Union[UuidValues, _Mapping]] = ..., int_values: _Optional[_Union[IntValues, _Mapping]] = ..., text_values: _Optional[_Union[TextValues, _Mapping]] = ...) -> None: ...

class NumberValues(_message.Message):
    __slots__ = ("values",)
    VALUES_FIELD_NUMBER: _ClassVar[int]
    values: bytes
    def __init__(self, values: _Optional[bytes] = ...) -> None: ...

class TextValues(_message.Message):
    __slots__ = ("values",)
    VALUES_FIELD_NUMBER: _ClassVar[int]
    values: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, values: _Optional[_Iterable[str]] = ...) -> None: ...

class BoolValues(_message.Message):
    __slots__ = ("values",)
    VALUES_FIELD_NUMBER: _ClassVar[int]
    values: _containers.RepeatedScalarFieldContainer[bool]
    def __init__(self, values: _Optional[_Iterable[bool]] = ...) -> None: ...

class ObjectValues(_message.Message):
    __slots__ = ("values",)
    VALUES_FIELD_NUMBER: _ClassVar[int]
    values: _containers.RepeatedCompositeFieldContainer[Properties]
    def __init__(self, values: _Optional[_Iterable[_Union[Properties, _Mapping]]] = ...) -> None: ...

class DateValues(_message.Message):
    __slots__ = ("values",)
    VALUES_FIELD_NUMBER: _ClassVar[int]
    values: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, values: _Optional[_Iterable[str]] = ...) -> None: ...

class UuidValues(_message.Message):
    __slots__ = ("values",)
    VALUES_FIELD_NUMBER: _ClassVar[int]
    values: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, values: _Optional[_Iterable[str]] = ...) -> None: ...

class IntValues(_message.Message):
    __slots__ = ("values",)
    VALUES_FIELD_NUMBER: _ClassVar[int]
    values: bytes
    def __init__(self, values: _Optional[bytes] = ...) -> None: ...

class GeoCoordinate(_message.Message):
    __slots__ = ("longitude", "latitude")
    LONGITUDE_FIELD_NUMBER: _ClassVar[int]
    LATITUDE_FIELD_NUMBER: _ClassVar[int]
    longitude: float
    latitude: float
    def __init__(self, longitude: _Optional[float] = ..., latitude: _Optional[float] = ...) -> None: ...

class PhoneNumber(_message.Message):
    __slots__ = ("country_code", "default_country", "input", "international_formatted", "national", "national_formatted", "valid")
    COUNTRY_CODE_FIELD_NUMBER: _ClassVar[int]
    DEFAULT_COUNTRY_FIELD_NUMBER: _ClassVar[int]
    INPUT_FIELD_NUMBER: _ClassVar[int]
    INTERNATIONAL_FORMATTED_FIELD_NUMBER: _ClassVar[int]
    NATIONAL_FIELD_NUMBER: _ClassVar[int]
    NATIONAL_FORMATTED_FIELD_NUMBER: _ClassVar[int]
    VALID_FIELD_NUMBER: _ClassVar[int]
    country_code: int
    default_country: str
    input: str
    international_formatted: str
    national: int
    national_formatted: str
    valid: bool
    def __init__(self, country_code: _Optional[int] = ..., default_country: _Optional[str] = ..., input: _Optional[str] = ..., international_formatted: _Optional[str] = ..., national: _Optional[int] = ..., national_formatted: _Optional[str] = ..., valid: bool = ...) -> None: ...



from typing import Dict, List, Optional, Sequence, Union

from typing_extensions import deprecated

from weaviate.connect.v4 import ConnectionAsync
from weaviate.rbac.models import (
    GroupAssignment,
    PermissionsInputType,
    PermissionsOutputType,
    Role,
    UserAssignment,
)

from .executor import _RolesExecutor

class _RolesAsync(_RolesExecutor[ConnectionAsync]):
    async def list_all(self) -> Dict[str, Role]: ...
    @deprecated(
        "This method is deprecated and will be removed in Q4 25. Please use `users.get_my_user()` instead."
    )
    async def get_current_roles(self) -> List[Role]: ...
    async def exists(self, role_name: str) -> bool: ...
    async def get(self, role_name: str) -> Optional[Role]: ...
    async def create(self, *, role_name: str, permissions: PermissionsInputType) -> Role: ...
    async def get_user_assignments(self, role_name: str) -> List[UserAssignment]: ...
    async def get_group_assignments(self, role_name: str) -> List[GroupAssignment]: ...
    @deprecated(
        "This method is deprecated and will be removed in Q4 25. Please use `roles.get_user_assignments` instead."
    )
    async def get_assigned_user_ids(self, role_name: str) -> List[str]: ...
    async def delete(self, role_name: str) -> None: ...
    async def add_permissions(
        self, *, permissions: PermissionsInputType, role_name: str
    ) -> None: ...
    async def remove_permissions(
        self, *, permissions: PermissionsInputType, role_name: str
    ) -> None: ...
    async def has_permissions(
        self,
        *,
        permissions: Union[
            PermissionsInputType, PermissionsOutputType, Sequence[PermissionsOutputType]
        ],
        role: str,
    ) -> bool: ...



from weaviate.classes.rbac import Permissions


def test_permissions_roles_only_manage_false() -> None:
    permissions = Permissions.roles(read=False, role="*")
    assert len(permissions) == 0


def test_permissions_roles_only_manage_true() -> None:
    permissions = Permissions.roles(read=True, role="*")
    assert len(permissions) == 1
    assert len(permissions[0].actions) == 1


def test_permissions_roles() -> None:
    permissions = Permissions.roles(read=True, create=False, role="*")
    assert len(permissions) == 1
    assert len(permissions[0].actions) == 1



import uuid
from dataclasses import dataclass
from enum import Enum
from typing import Generic, List, TypedDict, TypeVar, Union


class ReplicationType(str, Enum):
    """Enum for replication types."""

    COPY = "COPY"
    MOVE = "MOVE"


class ReplicateOperationState(str, Enum):
    """Enum for replication operation states."""

    REGISTERED = "REGISTERED"
    HYDRATING = "HYDRATING"
    FINALIZING = "FINALIZING"
    DEHYDRATING = "DEHYDRATING"
    READY = "READY"
    CANCELLED = "CANCELLED"


@dataclass
class ReplicateOperationStatus:
    """Class representing the status of a replication operation."""

    state: ReplicateOperationState
    errors: List[str]

    @classmethod
    def _from_weaviate(cls, data: dict) -> "ReplicateOperationStatus":
        return cls(
            state=ReplicateOperationState(data["state"]),
            errors=data["errors"] or [],
        )


H = TypeVar("H", None, List[ReplicateOperationStatus])


@dataclass
class _ReplicateOperation(Generic[H]):
    """Class representing a replication operation."""

    collection: str
    shard: str
    source_node: str
    status: ReplicateOperationStatus
    status_history: H
    target_node: str
    transfer_type: ReplicationType
    uuid: uuid.UUID

    @staticmethod
    def _from_weaviate(
        data: dict,
        include_history: bool,
    ):
        common = {
            "collection": data["collection"],
            "shard": data["shard"],
            "source_node": data["sourceNode"],
            "status": ReplicateOperationStatus._from_weaviate(data["status"]),
            "target_node": data["targetNode"],
            "transfer_type": ReplicationType(data["type"]),
            "uuid": uuid.UUID(data["id"]),
        }
        if include_history and data["statusHistory"] is not None:
            return _ReplicateOperation(
                **common,
                status_history=[
                    ReplicateOperationStatus._from_weaviate(status)
                    for status in data["statusHistory"]
                ],
            )
        return _ReplicateOperation(
            **common,
            status_history=None,
        )


ReplicateOperationWithoutHistory = _ReplicateOperation[None]
ReplicateOperationWithHistory = _ReplicateOperation[List[ReplicateOperationStatus]]

ReplicateOperation = Union[ReplicateOperationWithoutHistory, ReplicateOperationWithHistory]
ReplicateOperations = Union[
    List[ReplicateOperationWithoutHistory], List[ReplicateOperationWithHistory]
]


class _ReplicationShardReplicas(TypedDict):
    shard: str
    replicas: List[str]


class _ReplicationShardingState(TypedDict):
    collection: str
    shards: List[_ReplicationShardReplicas]


class _ReplicationShardingStateResponse(TypedDict):
    shardingState: _ReplicationShardingState


@dataclass
class ShardReplicas:
    """Class representing a shard replica."""

    name: str
    replicas: List[str]

    @staticmethod
    def _from_weaviate(data: _ReplicationShardReplicas):
        return ShardReplicas(
            name=data["shard"],
            replicas=data["replicas"],
        )


@dataclass
class ShardingState:
    """Class representing the sharding state of a collection."""

    collection: str
    shards: List[ShardReplicas]

    @staticmethod
    def _from_weaviate(data: _ReplicationShardingStateResponse):
        ss = data["shardingState"]
        return ShardingState(
            collection=ss["collection"],
            shards=[ShardReplicas._from_weaviate(shard) for shard in ss["shards"]],
        )
